import json
import time
from random import randint
import re
import warnings

from ..compat import (
    compat_urllib_error, compat_urllib_request,
    compat_http_client
)
from ..errors import ErrorHandler, ClientError, ClientConnectionError
from ..http import MultipartFormDataEncoder
from ..utils import (
    max_chunk_count_generator, max_chunk_size_generator,
    get_file_size
)
from ..compatpatch import ClientCompatPatch
from .common import ClientDeprecationWarning
from .common import MediaTypes
from socket import timeout, error as SocketError
from ssl import SSLError
try:
    ConnectionError = ConnectionError       # pylint: disable=redefined-builtin
except NameError:  # Python 2:
    class ConnectionError(Exception):
        pass


class MediaRatios(object):
    """
    Class holding valid aspect ratios (width: height) for media uploads.
    """

    # Based on IG sampling
    # and from https://help.instagram.com/1469029763400082
    #: Acceptable min, max values of with/height ratios for a standard media upload
    standard = 4.0 / 5.0, 90.0 / 47.0
    __device_ratios = [(3, 4), (2, 3), (5, 8), (3, 5), (9, 16), (10, 16), (40, 71)]
    __aspect_ratios = [1.0 * x[0] / x[1] for x in __device_ratios]

    #: Acceptable min, max values of with/height ratios for a story upload
    reel = min(__aspect_ratios), max(__aspect_ratios)


class UploadEndpointsMixin(object):
    """For endpoints relating to upload functionality."""

    EXTERNAL_LOC_SOURCES = {
        'foursquare': 'foursquare_v2_id',
        'facebook_places': 'facebook_places_id',
        'facebook_events': 'facebook_events_id'
    }

    def _validate_location(self, location):
        """
        Validates and patches a location dict for use with the upload functions

        :param location: dict containing location info
        :return:
        """
        location_keys = ['external_source', 'name', 'address']
        if not isinstance(location, dict):
            raise ValueError('Location must be a dict.')

        # patch location object returned from location_search
        if 'external_source' not in location and 'external_id_source' in location and 'external_id' in location:
            external_source = location['external_id_source']
            location['external_source'] = external_source
            if external_source in self.EXTERNAL_LOC_SOURCES:
                location[self.EXTERNAL_LOC_SOURCES[external_source]] = location['external_id']
        for k in location_keys:
            if not location.get(k):
                raise ValueError('Location dict must contain "{0!s}".'.format(k))
        for k, val in self.EXTERNAL_LOC_SOURCES.items():
            if location['external_source'] == k and not location.get(val):
                raise ValueError('Location dict must contain "{0!s}".'.format(val))

        media_loc = {
            'name': location['name'],
            'address': location['lat'],
            'external_source': location['external_source'],
        }
        if 'lat' in location and 'lng' in location:
            media_loc['lat'] = location['lat']
            media_loc['lng'] = location['lng']
        for k, val in self.EXTERNAL_LOC_SOURCES.items():
            if location['external_source'] == k:
                media_loc['external_source'] = k
                media_loc[val] = location[val]
        return media_loc

    @staticmethod
    def standard_ratios():  # pragma: no cover
        """
        Deprecated. Use MediaRatios.standard instead.
        Acceptable min, max values of with/height ratios for a standard media upload

        :return: tuple of (min. ratio, max. ratio)
        """
        warnings.warn(
            'Client.standard_ratios() is deprecated. '
            'Please use MediaRatios.standard instead.',
            ClientDeprecationWarning
        )
        return MediaRatios.standard

    @staticmethod
    def reel_ratios():  # pragma: no cover
        """
        Deprecated. Use MediaRatios.reel instead.
        Acceptable min, max values of with/height ratios for a story upload

        :return: tuple of (min. ratio, max. ratio)
        """
        warnings.warn(
            'Client.reel_ratios() is deprecated. '
            'Please use MediaRatios.reel instead.',
            ClientDeprecationWarning
        )
        return MediaRatios.reel

    @classmethod
    def compatible_aspect_ratio(cls, size):
        """
        Helper method to check aspect ratio for standard uploads

        :param size: tuple of (width, height)
        :return: True/False
        """
        min_ratio, max_ratio = MediaRatios.standard
        width, height = size
        this_ratio = 1.0 * width / height
        return min_ratio <= this_ratio <= max_ratio

    @classmethod
    def reel_compatible_aspect_ratio(cls, size):
        """
        Helper method to check aspect ratio for story uploads

        :param size: tuple of (width, height)
        :return: True/False
        """
        min_ratio, max_ratio = MediaRatios.reel
        width, height = size
        this_ratio = 1.0 * width / height
        return min_ratio <= this_ratio <= max_ratio

    def configure(self, upload_id, size, caption='', location=None,
                  disable_comments=False, is_sidecar=False):
        """
        Finalises a photo upload. This should not be called directly.
        Use :meth:`post_photo` instead.

        :param upload_id:
        :param size: tuple of (width, height)
        :param caption:
        :param location: a dict of venue/location information,
                         from :meth:`location_search` or :meth:`location_fb_search`
        :param disable_comments:
        :param is_sidecar: bool flag for album upload
        :return:
        """
        if not self.compatible_aspect_ratio(size):
            raise ValueError('Incompatible aspect ratio.')

        endpoint = 'media/configure/'
        width, height = size
        params = {
            'caption': caption,
            'media_folder': 'Instagram',
            'source_type': '4',
            'upload_id': upload_id,
            'device': {
                'manufacturer': self.phone_manufacturer,
                'model': self.phone_device,
                'android_version': self.android_version,
                'android_release': self.android_release
            },
            'edits': {
                'crop_original_size': [width * 1.0, height * 1.0],
                'crop_center': [0.0, -0.0],
                'crop_zoom': 1.0
            },
            'extra': {
                'source_width': width,
                'source_height': height,
            }
        }
        if location:
            media_loc = self._validate_location(location)
            params['location'] = json.dumps(media_loc)
            if 'lat' in location and 'lng' in location:
                params['geotag_enabled'] = '1'
                params['exif_latitude'] = '0.0'
                params['exif_longitude'] = '0.0'
                params['posting_latitude'] = str(location['lat'])
                params['posting_longitude'] = str(location['lng'])
                params['media_latitude'] = str(location['lat'])
                params['media_latitude'] = str(location['lng'])
        if disable_comments:
            params['disable_comments'] = '1'

        if is_sidecar:
            return params

        params.update(self.authenticated_params)
        res = self._call_api(endpoint, params=params)
        if self.auto_patch and res.get('media'):
            ClientCompatPatch.media(res.get('media'), drop_incompat_keys=self.drop_incompat_keys)
        return res

    def configure_video(self, upload_id, size, duration, thumbnail_data, caption='',
                        location=None, disable_comments=False, is_sidecar=False):
        """
        Finalises a video upload. This should not be called directly.
        Use :meth:`post_video` instead.

        :param upload_id:
        :param size: tuple of (width, height)
        :param duration: in seconds
        :param thumbnail_data: byte string of thumbnail photo
        :param caption:
        :param location: a dict of venue/location information,
                         from :meth:`location_search` or :meth:`location_fb_search`
        :param disable_comments:
        :param is_sidecar: bool flag for album upload
        :return:
        """
        if not self.compatible_aspect_ratio(size):
            raise ValueError('Incompatible aspect ratio.')

        # upload video thumbnail
        self.post_photo(thumbnail_data, size, caption, upload_id, location=location,
                        disable_comments=disable_comments, is_sidecar=is_sidecar)

        width, height = size
        params = {
            'upload_id': upload_id,
            'caption': caption,
            'source_type': '3',
            'poster_frame_index': 0,
            'length': duration * 1.0,
            'audio_muted': False,
            'filter_type': '0',
            'video_result': 'deprecated',
            'clips': {
                'length': duration * 1.0,
                'source_type': '3',
                'camera_position': 'back'
            },
            'device': {
                'manufacturer': self.phone_manufacturer,
                'model': self.phone_device,
                'android_version': self.android_version,
                'android_release': self.android_release
            },
            'extra': {
                'source_width': width,
                'source_height': height
            }
        }
        if disable_comments:
            params['disable_comments'] = '1'
        if location:
            media_loc = self._validate_location(location)
            params['location'] = json.dumps(media_loc)
            if 'lat' in location and 'lng' in location:
                params['geotag_enabled'] = '1'
                params['av_latitude'] = '0.0'
                params['av_longitude'] = '0.0'
                params['posting_latitude'] = str(location['lat'])
                params['posting_longitude'] = str(location['lng'])
                params['media_latitude'] = str(location['lat'])
                params['media_latitude'] = str(location['lng'])

        if is_sidecar:
            return params

        params.update(self.authenticated_params)
        res = self._call_api('media/configure/', params=params, query={'video': 1})
        if res.get('media') and self.auto_patch:
            ClientCompatPatch.media(res.get('media'), drop_incompat_keys=self.drop_incompat_keys)
        return res

    def configure_to_reel(self, upload_id, size):
        """
        Finalises a photo story upload. This should not be called directly.
        Use :meth:`post_photo_story` instead.

        :param upload_id:
        :param size: tuple of (width, height)
        :return:
        """
        if not self.reel_compatible_aspect_ratio(size):
            raise ValueError('Incompatible aspect ratio.')

        endpoint = 'media/configure_to_story/'
        width, height = size
        params = {
            'source_type': '4',
            'upload_id': upload_id,
            'story_media_creation_date': str(int(time.time()) - randint(11, 20)),
            'client_shared_at': str(int(time.time()) - randint(3, 10)),
            'client_timestamp': str(int(time.time())),
            'configure_mode': 1,      # 1 - REEL_SHARE, 2 - DIRECT_STORY_SHARE
            'device': {
                'manufacturer': self.phone_manufacturer,
                'model': self.phone_device,
                'android_version': self.android_version,
                'android_release': self.android_release
            },
            'edits': {
                'crop_original_size': [width * 1.0, height * 1.0],
                'crop_center': [0.0, 0.0],
                'crop_zoom': 1.3333334
            },
            'extra': {
                'source_width': width,
                'source_height': height,
            }
        }
        params.update(self.authenticated_params)
        res = self._call_api(endpoint, params=params)
        if self.auto_patch and res.get('media'):
            ClientCompatPatch.media(res.get('media'), drop_incompat_keys=self.drop_incompat_keys)
        return res

    def configure_video_to_reel(self, upload_id, size, duration, thumbnail_data):
        """
        Finalises a video story upload. This should not be called directly.
        Use :meth:`post_video_story` instead.

        :param upload_id:
        :param size: tuple of (width, height)
        :param duration: in seconds
        :param thumbnail_data: byte string of thumbnail photo
        :return:
        """
        if not self.reel_compatible_aspect_ratio(size):
            raise ValueError('Incompatible aspect ratio.')

        res = self.post_photo(thumbnail_data, size, '', upload_id=upload_id, to_reel=True)

        width, height = size
        params = {
            'source_type': '4',
            'upload_id': upload_id,
            'story_media_creation_date': str(int(time.time()) - randint(11, 20)),
            'client_shared_at': str(int(time.time()) - randint(3, 10)),
            'client_timestamp': str(int(time.time())),
            'configure_mode': 1,      # 1 - REEL_SHARE, 2 - DIRECT_STORY_SHARE
            'poster_frame_index': 0,
            'length': duration * 1.0,
            'audio_muted': False,
            'filter_type': '0',
            'video_result': 'deprecated',
            'clips': {
                'length': duration * 1.0,
                'source_type': '4',
                'camera_position': 'back'
            },
            'device': {
                'manufacturer': self.phone_manufacturer,
                'model': self.phone_device,
                'android_version': self.android_version,
                'android_release': self.android_release
            },
            'extra': {
                'source_width': width,
                'source_height': height,
            },
        }

        params.update(self.authenticated_params)
        res = self._call_api('media/configure_to_story/', params=params, query={'video': '1'})
        if self.auto_patch and res.get('media'):
            ClientCompatPatch.media(res.get('media'), drop_incompat_keys=self.drop_incompat_keys)
        return res

    def post_photo(self, photo_data, size, caption='', upload_id=None, to_reel=False, **kwargs):
        """
        Upload a photo.

        [CAUTION] FLAKY, IG is very finicky about sizes, etc, needs testing.

        :param photo_data: byte string of the image
        :param size: tuple of (width, height)
        :param caption:
        :param upload_id:
        :param to_reel: a Story photo
        :param kwargs:
            - **location**: a dict of venue/location information, from :meth:`location_search`
              or :meth:`location_fb_search`
            - **disable_comments**: bool to disable comments
        :return:
        """
        warnings.warn('This endpoint has not been fully tested.', UserWarning)

        # if upload_id is provided, it's a thumbnail for a vid upload
        for_video = True if upload_id else False

        if not for_video:
            if not to_reel and not self.compatible_aspect_ratio(size):
                raise ValueError('Incompatible aspect ratio.')
            if to_reel and not self.reel_compatible_aspect_ratio(size):
                raise ValueError('Incompatible reel aspect ratio.')
            if not 320 <= size[0] <= 1080:
                # range from https://help.instagram.com/1631821640426723
                raise ValueError('Invalid image width.')

        location = kwargs.pop('location', None)
        if location:
            self._validate_location(location)
        disable_comments = True if kwargs.pop('disable_comments', False) else False

        is_sidecar = kwargs.pop('is_sidecar', False)
        if not upload_id:
            upload_id = str(int(time.time() * 1000))

        endpoint = 'upload/photo/'
        fields = [
            ('upload_id', upload_id),
            ('_uuid', self.uuid),
            ('_csrftoken', self.csrftoken),
            ('image_compression', '{"lib_name":"jt","lib_version":"1.3.0","quality":"87"}')
        ]
        if is_sidecar:
            fields.append(('is_sidecar', '1'))
            if for_video:
                fields.append(('media_type', MediaTypes.VIDEO))

        files = [
            ('photo', 'pending_media_{0!s}{1!s}'.format(str(int(time.time() * 1000)), '.jpg'),
             'application/octet-stream', photo_data)
        ]

        content_type, body = MultipartFormDataEncoder().encode(fields, files)
        headers = self.default_headers
        headers['Content-Type'] = content_type
        headers['Content-Length'] = len(body)

        endpoint_url = '{0}{1}'.format(self.api_url.format(version='v1'), endpoint)
        req = compat_urllib_request.Request(endpoint_url, body, headers=headers)
        try:
            self.logger.debug('POST {0!s}'.format(endpoint_url))
            response = self.opener.open(req, timeout=self.timeout)
        except compat_urllib_error.HTTPError as e:
            error_response = self._read_response(e)
            self.logger.debug('RESPONSE: {0:d} {1!s}'.format(e.code, error_response))
            ErrorHandler.process(e, error_response)
        except (SSLError, timeout, SocketError,
                compat_urllib_error.URLError,   # URLError is base of HTTPError
                compat_http_client.HTTPException) as connection_error:
            raise ClientConnectionError('{} {}'.format(
                connection_error.__class__.__name__, str(connection_error)))

        post_response = self._read_response(response)
        self.logger.debug('RESPONSE: {0:d} {1!s}'.format(response.code, post_response))
        json_response = json.loads(post_response)

        if for_video and is_sidecar:
            return json_response

        upload_id = json_response['upload_id']

        # # NOTES: Logging traffic doesn't seem to indicate any additional "configure" after upload
        # # BUT not doing a "configure" causes a video post to fail with a
        # # "Other media configure error: b'yEmZkUpAj4'" error
        # if for_video:
        #     logger.debug('Skip photo configure.')
        #     return json_response
        if to_reel:
            return self.configure_to_reel(upload_id, size)
        else:
            return self.configure(upload_id, size, caption=caption, location=location,
                                  disable_comments=disable_comments, is_sidecar=is_sidecar)

    def post_video(self, video_data, size, duration, thumbnail_data, caption='', to_reel=False, **kwargs):
        """
        Upload a video

        [CAUTION] FLAKY, IG is very picky about sizes, etc, needs testing.

        :param video_data: byte string or a file-like object of the video content
        :param size: tuple of (width, height)
        :param duration: in seconds
        :param thumbnail_data: byte string of the video thumbnail content
        :param caption:
        :param to_reel: post to reel as Story
        :param kwargs:
             - **location**: a dict of venue/location information, from :meth:`location_search`
               or :meth:`location_fb_search`
             - **disable_comments**: bool to disable comments
             - **max_retry_count**: maximum attempts to reupload. Default 10.
        :return:
        """
        warnings.warn('This endpoint has not been fully tested.', UserWarning)

        if not to_reel and not self.compatible_aspect_ratio(size):
            raise ValueError('Incompatible aspect ratio.')

        if to_reel and not self.reel_compatible_aspect_ratio(size):
            raise ValueError('Incompatible reel aspect ratio.')

        if not 612 <= size[0] <= 1080:
            # range was determined through sampling of video uploads
            raise ValueError('Invalid video width.')

        if duration < 3.0:
            raise ValueError('Duration is less than 3s.')

        if not to_reel and duration > 60.0:
            raise ValueError('Duration is more than 60s.')

        if to_reel and duration > 15.0:
            raise ValueError('Duration is more than 15s.')

        max_file_len = 50 * 1024 * 1000
        try:
            video_file_len = len(video_data)
        except TypeError:
            video_file_len = get_file_size(video_data)
        if video_file_len > max_file_len:
            raise ValueError('Video file is too big.')

        location = kwargs.pop('location', None)
        if location:
            self._validate_location(location)
        disable_comments = True if kwargs.pop('disable_comments', False) else False

        endpoint = 'upload/video/'
        upload_id = str(int(time.time() * 1000))

        width, height = size
        params = {
            '_csrftoken': self.csrftoken,
            '_uuid': self.uuid,
            'upload_id': upload_id,
        }
        is_sidecar = kwargs.pop('is_sidecar', False)
        if is_sidecar:
            params['is_sidecar'] = '1'
        else:
            params.update({
                'media_type': MediaTypes.VIDEO,
                'upload_media_duration_ms': int(duration * 1000),
                'upload_media_width': width,
                'upload_media_height': height
            })

        res = self._call_api(endpoint, params=params, unsigned=True)
        upload_url = res['video_upload_urls'][-1]['url']
        upload_job = res['video_upload_urls'][-1]['job']

        successful_chunk_ranges = []
        all_done = False

        max_retry_count = kwargs.pop('max_retry_count', 10)
        configure_delay = 0
        for _ in range(max_retry_count + 1):

            try:
                # Prevent excessively small chunks
                if video_file_len > 1 * 1024 * 1000:
                    # max num of chunks = 4
                    chunk_generator = max_chunk_count_generator(4, video_data)
                else:
                    # max chunk size = 350,000 so that we'll always have
                    # <4 chunks when it's <1mb
                    chunk_generator = max_chunk_size_generator(350000, video_data)

                for chunk, data in chunk_generator:
                    skip_chunk = False
                    for received_chunk in successful_chunk_ranges:
                        if received_chunk[0] <= chunk.start and received_chunk[1] >= (chunk.end - 1):
                            skip_chunk = True
                            break
                    if skip_chunk:
                        self.logger.debug('Skipped chunk: {0:d} - {1:d}'.format(chunk.start, chunk.end - 1))
                        continue

                    headers = self.default_headers
                    headers['Connection'] = 'keep-alive'
                    headers['Content-Type'] = 'application/octet-stream'
                    headers['Content-Disposition'] = 'attachment; filename="video.mov"'
                    headers['Session-ID'] = upload_id
                    if is_sidecar:
                        headers['Cookie'] = 'sessionid=' + self.get_cookie_value('sessionid')
                    headers['job'] = upload_job
                    headers['Content-Length'] = chunk.length
                    headers['Content-Range'] = 'bytes {0:d}-{1:d}/{2:d}'.format(
                        chunk.start, chunk.end - 1, video_file_len)
                    self.logger.debug('POST {0!s}'.format(upload_url))
                    self.logger.debug('Uploading Content-Range: {0!s}'.format(headers['Content-Range']))

                    req = compat_urllib_request.Request(
                        str(upload_url), data=data, headers=headers)

                    try:
                        res = self.opener.open(req, timeout=self.timeout)
                        post_response = self._read_response(res)
                        self.logger.debug('RESPONSE: {0:d} {1!s}'.format(res.code, post_response))
                        if res.info().get('Content-Type', '').startswith('application/json'):
                            # last chunk
                            upload_res = json.loads(post_response)
                            configure_delay = int(upload_res.get('configure_delay_ms', 0)) / 1000.0
                            all_done = True
                            break
                        else:
                            successful_chunk_ranges = []
                            post_progress = post_response.split(',')
                            for progress in post_progress:
                                mobj = re.match(r'(?P<start>[0-9]+)\-(?P<end>[0-9]+)/(?P<total>[0-9]+)', progress)
                                if mobj:
                                    successful_chunk_ranges.append((int(mobj.group('start')), int(mobj.group('end'))))
                                else:
                                    self.logger.error(
                                        'Received unexpected chunk upload response: {0!s}'.format(post_response))
                                    raise ClientError(
                                        'Upload has failed due to unexpected upload response: {0!s}'.format(
                                            post_response),
                                        code=500)

                    except compat_urllib_error.HTTPError as e:
                        error_response = self._read_response(e)
                        self.logger.debug('RESPONSE: {0:d} {1!s}'.format(e.code, error_response))
                        ErrorHandler.process(e, error_response)

                    except (SSLError, timeout, SocketError,
                            compat_urllib_error.URLError,   # URLError is base of HTTPError
                            compat_http_client.HTTPException) as connection_error:
                        raise ClientConnectionError('{} {}'.format(
                            connection_error.__class__.__name__, str(connection_error)))

                else:
                    # if not break due to completed chunks then continue with next chunk
                    continue
                break

            except ClientConnectionError as cce:
                # connectivity prob, so let's just continue with another attempt
                self.logger.warning('ClientConnectionError posting chunks: {}'.format(
                    str(cce)))
                continue

        if not all_done:
            raise ClientError('Upload has failed due to incomplete chunk uploads.', code=500)

        if not configure_delay:
            configure_delay = 3
        configure_retry_max = 2

        for i in range(1, configure_retry_max + 1):
            try:
                if not to_reel:
                    result = self.configure_video(
                        upload_id, size, duration, thumbnail_data, caption=caption, location=location,
                        disable_comments=disable_comments, is_sidecar=is_sidecar)
                else:
                    result = self.configure_video_to_reel(
                        upload_id, size, duration, thumbnail_data)
                return result
            except ClientConnectionError as cce:
                if i < configure_retry_max:
                    self.logger.warning(
                        'Retry configure after {0:f} seconds: {1:s}'.format(
                            configure_delay, cce.msg))
                    time.sleep(configure_delay)
                else:
                    raise
            except ClientError as ce:
                if (ce.code == 202 or ce.msg == 'Transcode timeout') and i < configure_retry_max:
                    self.logger.warning('Retry configure after {0:f} seconds'.format(configure_delay))
                    time.sleep(configure_delay)
                else:
                    raise

    def post_photo_story(self, photo_data, size):
        """
        Upload a photo story

        :param photo_data: byte string of the image
        :param size: tuple of (width, height)
        :return:
        """
        return self.post_photo(
            photo_data=photo_data, size=size, to_reel=True)

    def post_video_story(self, video_data, size, duration, thumbnail_data):
        """
        Upload a video story

        :param video_data: byte string or a file-like object of the video content
        :param size: tuple of (width, height)
        :param duration: in seconds
        :param thumbnail_data: byte string of the video thumbnail content
        :return:
        """
        return self.post_video(
            video_data=video_data, size=size, duration=duration,
            thumbnail_data=thumbnail_data, to_reel=True)

    def post_album(self, medias, caption='', location=None, **kwargs):
        """
        Post an album of up to 10 photos/videos.

        :param medias: an iterable list/collection of media dict objects

            .. code-block:: javascript

                medias = [
                    {"type": "image", "size": (720, 720), "data": "..."},
                    {
                        "type": "image", "size": (720, 720),
                        "usertags": [{"user_id":4292127751, "position":[0.625347,0.4384531]}],
                        "data": "..."
                    },
                    {"type": "video", "size": (720, 720), "duration": 12.4, "thumbnail": "...", "data": "..."}
                ]

        :param caption:
        :param location:
        :return:
        """
        album_upload_id = str(int(time.time() * 1000))
        children_metadata = []
        for media in medias:
            if len(children_metadata) >= 10:
                continue
            if media.get('type', '') not in ['image', 'video']:
                raise ValueError('Invalid media type: {0!s}'.format(media.get('type', '')))
            if not media.get('data'):
                raise ValueError('Data not specified.')
            if not media.get('size'):
                raise ValueError('Size not specified.')
            if media['type'] == 'video':
                if not media.get('duration'):
                    raise ValueError('Duration not specified.')
                if not media.get('thumbnail'):
                    raise ValueError('Thumbnail not specified.')
            aspect_ratio = (media['size'][0] * 1.0) / (media['size'][1] * 1.0)
            if aspect_ratio > 1.0 or aspect_ratio < 1.0:
                raise ValueError('Invalid media aspect ratio.')

            if media['type'] == 'video':
                metadata = self.post_video(
                    video_data=media['data'],
                    size=media['size'],
                    duration=media['duration'],
                    thumbnail_data=media['thumbnail'],
                    is_sidecar=True
                )
            else:
                metadata = self.post_photo(
                    photo_data=media['data'],
                    size=media['size'],
                    is_sidecar=True,
                )
                if media.get('usertags'):
                    usertags = media['usertags']
                    utags = {'in': [{'user_id': str(u['user_id']), 'position': u['position']} for u in usertags]}
                    metadata['usertags'] = json.dumps(utags, separators=(',', ':'))
            children_metadata.append(metadata)

        if len(children_metadata) <= 1:
            raise ValueError('Invalid number of media objects: {0:d}'.format(len(children_metadata)))

        # configure as sidecar
        endpoint = 'media/configure_sidecar/'
        params = {
            'caption': caption,
            'client_sidecar_id': album_upload_id,
            'children_metadata': children_metadata
        }
        if location:
            media_loc = self._validate_location(location)
            params['location'] = json.dumps(media_loc)
            if 'lat' in location and 'lng' in location:
                params['geotag_enabled'] = '1'
                params['exif_latitude'] = '0.0'
                params['exif_longitude'] = '0.0'
                params['posting_latitude'] = str(location['lat'])
                params['posting_longitude'] = str(location['lng'])
                params['media_latitude'] = str(location['lat'])
                params['media_latitude'] = str(location['lng'])
        disable_comments = kwargs.pop('disable_comments', False)
        if disable_comments:
            params['disable_comments'] = '1'

        params.update(self.authenticated_params)
        res = self._call_api(endpoint, params=params)
        if self.auto_patch and res.get('media'):
            ClientCompatPatch.media(res.get('media'), drop_incompat_keys=self.drop_incompat_keys)
        return res



from .bot import Bot
from .execute import execute
import asyncio
import atexit
import concurrent.futures.thread




async def async_execute(script, variables={}, Executor=concurrent.futures.ThreadPoolExecutor(max_workers=5)):
    loop = asyncio.get_event_loop()
    atexit.unregister(concurrent.futures.thread._python_exit)
    # Executor.shutdown = lambda wait: None
    return await loop.run_in_executor(Executor, lambda: execute(script, variables))





from .common import decorate
from ..nodes import Node, node_classes

from .common import dotdict
from ..bot import Bot
from funcy import ignore
import json
from itertools import islice
from colorama import init, Fore




@decorate(accepts=(*node_classes.values(),), returns=Node)
def _print(bot: Bot, nodes,  args):


    def process(node):
        """
        model:
            name:      x.full_name
            id:        x.pk
            followers: x.followers_count
        """
        assert isinstance(args, str)
        value = evaluate(args, node, bot=bot)

        init()
        print()
        print(Fore.CYAN + json.dumps(value, indent=4))
        print()
        return node

    nodes = map(process, nodes)

    return nodes, {}



def evaluate(expr, node, bot):
    x = node
    return xeval(expr, x)


def xeval(expr, x):
    try:
        return eval(expr, dotdict(x=x))

    except (KeyError, AttributeError):
        return None



from .nodes_edges import nodes_edges
from .make_bot import make_bot
from .populate import populate_object, populate_string
from .assert_good_script import assert_good_script
from .reducer import  reducer
from .support import dotdict, merge
from collections import deque
from functools import reduce
import time
import traceback
import os
import re
import yaml
from .notification_events import notification_events
from .direct_inbox_events import direct_inbox_events
from .errors import (
    ClientError,
    ClientLoginRequiredError,
    ClientCookieExpiredError,
    ClientConnectionError,
    ClientThrottledError,
    ClientLoginError,
    ClientReqHeadersTooLargeError,
    ClientCheckpointRequiredError,
    ClientChallengeRequiredError,
    ClientSentryBlockError,
    Stop,
)

yaml.reader.Reader.NON_PRINTABLE = re.compile(
    u'[^\x09\x0A\x0D\x20-\x7E\x85\xA0-\uD7FF\uE000-\uFFFD\U00010000-\U0010FFFF]')


DEBUG = bool(os.environ.get('DEBUG'))

def execute(script_string, variables={}) -> [dict]:

    if "---" in script_string:
        scripts = script_string.split('---')
        def _reducer(acc, script):
            nonlocal variables
            variables.update(acc)
            return merge(acc, execute(script, variables))
        return reduce(_reducer, scripts, {})


    script = obj_from_yaml(script_string, variables)

    if not script:
        return { 'events': [] }

    assert_good_script(script)

    try:
        bot = make_bot(script, variables)
        
    except ClientCheckpointRequiredError as e:
        print(str(e))
        raise e from None
        
    except ClientSentryBlockError as e:
        print(str(e))
        raise e from None

    except ClientError as e:
        print(str(e))
        raise e from None



    script_name = script['name'] if 'name' in script else None
    if not script_name:
        bot.logger.warning('no script name')    
    script_name = script_name or 'not named script'
    
    bot.logger.info(f'# SCRIPT {script_name}')

    result = deque()

    try:
        for action in script['actions']:
            action_name = action['name'] if 'name' in action else 'unnmaed action'
            bot.logger.info(f'# ACTION {action_name}')

            nodes, edges = nodes_edges(action, bot)
            start_data = { 'events': list(notification_events(bot)) + list(direct_inbox_events(bot)) }

            begin_state = dotdict(nodes=(node for node in nodes), bot=bot, data=deque([start_data]), errors=[])
            bot.logger.info(f'# with nodes {nodes}')

            end_state = reduce(reducer, edges, begin_state)
            result.extend(end_state['data'])

    except (KeyboardInterrupt, SystemExit):
        bot.logger.warning('keyboard interrupt')
        raise

    except Exception as exc:
        print(
            exc.__class__.__name__,
            ':',
            exc,
            '\n',
            '\n'.join(traceback.format_exc().split('\n'))
        )
        raise

    else:
        result = dict(reduce(merge, result))
        return result



def obj_from_yaml(script, variables):
    if isinstance(script, str):
        script = populate_string(script, variables)
        try:
            data = yaml.safe_load(script)
            if not isinstance(data, dict):
                return {}
            return data
        except Exception:
            return {}
    else:
        return populate_object(script, variables)





    # try:
    #     for action in script['actions']:
    #
    #
    #
    #         threads = []
    #         task = make_task(action)
    #         name = task['name']
    #
    #         for (task, bot) in partitionate(task, bots):
    #             state = dict(nodes=task.nodes, bot=bot, data=dict(), errors=[])
    #             threads += [Reducer(state, task.edges)]
    #             # bot.logger.debug('edges : {}, nodes: {}'.format(edges, list(state['nodes'])))
    #
    #
    #         threads = start(threads)
    #         threads = wait(threads)
    #
    #         data['__' + interaction + '_interaction__'] = [thread.get_data() for thread in threads]
    #         # {'thread' + thread.name: thread.get_data() for thread in threads}



import os
import traceback
from ..nodes import Node, node_classes
from .common import decorate
from ..api.instagram_private_api.errors import (
    ClientError,
    ClientLoginRequiredError,
    ClientCookieExpiredError,
    ClientConnectionError,
    ClientThrottledError,
    ClientReqHeadersTooLargeError,
)
from ..bot import Bot
from ..handle_errors import handle



DEBUG = bool(os.environ.get('DEBUG'))

@decorate(accepts=(*node_classes.values(),), returns=Node)
def evaluate(bot: Bot, nodes,  args) -> Node:

    total = args['info']['total_nodes']
    count = 0
    nodes = iter(nodes)


    while True:
        try:
            res = handle(lambda: next(nodes), bot)
        except StopIteration:
            break
        except Exception as e:
            # bot.logger.error(f'unexpected exception {e}')
            raise e from None
        else:
            count += 1
            percentage = int(count / ((total*100) or 1))
            percentage = (str(percentage) + '%').center(5)
            bot.logger.info( f'{percentage}:  {count} nodes out of {total}')

        


    # bot.logger.warning(nodes[:3])

    return nodes, {}




# try:
#             next(nodes)

#         except StopIteration:
#             break

#         except ClientLoginRequiredError as e:
#             bot.logger.error(str(e))
#             bot.relogin()

#         except ClientConnectionError as e:
#             bot.logger.error(str(e))

#         except (ClientReqHeadersTooLargeError, ClientThrottledError) as e:
#             bot.logger.error(str(e))
#             bot.sleep(5 * 60)

#         except ClientError as e: # when trying to see private user
#             bot.logger.error(str(e))
#             bot.sleep()

#         except Exception as e:
#             bot.logger.error('unexpected exception {e}')
#             raise e from None



import json
from colorama import init, Fore
from .execute import execute
# from .debug import unmask



"""
python -m src /etc/template.yml --data-file /etc/data.json
"""

import sys



def load(file):
    with open(file, 'r') as f:
        return f.read()

if __name__ == '__main__':
    template_path = sys.argv[-1]
    template = load(template_path)
    out = execute(template)
    print(Fore.CYAN + json.dumps(out, indent=4))




import json
import time

from ..utils import raise_if_invalid_rank_token
from ..compatpatch import ClientCompatPatch


class LocationsEndpointsMixin(object):
    """For endpoints related to location functionality."""

    def location_info(self, location_id):
        """
        Get a location info

        :param location_id:
        :return:
            .. code-block:: javascript

                {
                  "status": "ok",
                  "location": {
                    "external_source": "facebook_places",
                    "city": "",
                    "name": "Berlin Brandenburger Tor",
                    "facebook_places_id": 114849465334163,
                    "address": "Pariser Platz",
                    "lat": 52.51588,
                    "pk": 229573811,
                    "lng": 13.37892
                  }
                }
        """
        endpoint = 'locations/{location_id!s}/info/'.format(**{'location_id': location_id})
        return self._call_api(endpoint)

    def location_related(self, location_id, **kwargs):
        """
        Get related locations

        :param location_id:
        :return:
        """
        endpoint = 'locations/{location_id!s}/related/'.format(**{'location_id': location_id})
        query = {
            'visited': json.dumps([{'id': location_id, 'type': 'location'}], separators=(',', ':')),
            'related_types': json.dumps(['location'], separators=(',', ':'))}
        query.update(kwargs)
        return self._call_api(endpoint, query=query)

    def location_search(self, latitude, longitude, query=None, **kwargs):
        """
        Location search

        :param latitude:
        :param longitude:
        :param query:
        :return:
        """
        query_params = {
            'rank_token': self.rank_token,
            'latitude': latitude,
            'longitude': longitude,
            'timestamp': int(time.time())
        }
        if query:
            query_params['search_query'] = query
        query_params.update(kwargs)
        return self._call_api('location_search/', query=query_params)

    def location_fb_search(self, query, rank_token, exclude_list=[], **kwargs):
        """
        Search for locations by query text

        :param query: search terms
        :param rank_token: Required for paging through a single feed. See examples/pagination.py
        :param exclude_list: List of numerical location IDs to exclude
        :param kwargs:
        :return:
        """
        raise_if_invalid_rank_token(rank_token)

        if not exclude_list:
            exclude_list = []

        query_params = {
            'query': query,
            'timezone_offset': self.timezone_offset,
            'count': 30,
            'exclude_list': json.dumps(exclude_list, separators=(',', ':')),
            'rank_token': rank_token,
        }
        query_params.update(kwargs)
        return self._call_api('fbsearch/places/', query=query_params)

    def location_section(self, location_id, rank_token, tab='ranked', **kwargs):
        """
        Get a location feed

        :param location_id:
        :param rank_token: Required for paging through a single feed and can be generated with
            :meth:`generate_uuid`. You should use the same rank_token for paging through a single location.
        :param tab: One of 'ranked', 'recent'
        :kwargs:
            **extract**: return the array of media items only
            **page**: for pagination
            **next_media_ids**: array of media_id (int) for pagination
            **max_id**: for pagination
        :return:
        """
        raise_if_invalid_rank_token(rank_token)
        if tab not in ('ranked', 'recent'):
            raise ValueError('Invalid tab: {}'.format(tab))

        extract_media_only = kwargs.pop('extract', False)
        endpoint = 'locations/{location_id!s}/sections/'.format(**{'location_id': location_id})
        params = {
            'rank_token': rank_token,
            'tab': tab,
            'session_id': self.session_id,
        }

        # explicitly set known paging parameters to avoid triggering server-side errors
        if kwargs.get('max_id'):
            params['max_id'] = kwargs.pop('max_id')
        if kwargs.get('page'):
            params['page'] = kwargs.pop('page')
        if kwargs.get('next_media_ids'):
            params['next_media_ids'] = json.dumps(kwargs.pop('next_media_ids'), separators=(',', ':'))
        kwargs.pop('max_id', None)
        kwargs.pop('page', None)
        kwargs.pop('next_media_ids', None)

        params.update(kwargs)
        results = self._call_api(endpoint, params=params, unsigned=True)
        extracted_medias = []
        if self.auto_patch:
            for s in results.get('sections', []):
                for m in s.get('layout_content', {}).get('medias', []):
                    if m.get('media'):
                        ClientCompatPatch.media(m['media'], drop_incompat_keys=self.drop_incompat_keys)
                        if extract_media_only:
                            extracted_medias.append(m['media'])
        if extract_media_only:
            return extracted_medias
        return results

    def location_stories(self, location_id, **kwargs):
        """
        Get a location story feed

        :param location_id:
        :param rank_token: Required for paging through a single feed and can be generated with
            :meth:`generate_uuid`. You should use the same rank_token for paging through a single location.
        :return:
        """
        endpoint = 'locations/{location_id!s}/story/'.format(**{'location_id': location_id})
        # params = {
        #     'rank_token': rank_token,
        #     'tab': tab,
        #     'session_id': self.session_id,
        # }
        # params.update(kwargs)
        # return self._call_api(endpoint, params=params)
        return self._call_api(endpoint)



# -*- coding: utf-8 -*-
from .endpoints.common import MediaTypes


class ClientCompatPatch(object):
    """Utility to make entities from the private api similar to the ones
    from the public one by adding the necessary properties, and if required,
    remove any incompatible properties (to save storage space for example).
    """
    FILTERS = {
        -2: 'OES',
        -1: 'YUV',
        0: 'Normal',
        1: 'X-Pro II',
        2: 'Lo-Fi',
        3: 'Earlybird',
        10: 'Inkwell',
        14: '1977',
        15: 'Nashville',
        16: 'Kelvin',
        17: 'Mayfair',
        18: 'Sutro',
        19: 'Toaster',
        20: 'Walden',
        21: 'Hefe',
        22: 'Brannan',
        23: 'Rise',
        24: 'Amaro',
        25: 'Valencia',
        26: 'Hudson',
        27: 'Sierra',
        28: 'Willow',
        105: 'Dogpatch',
        106: 'Vesper',
        107: 'Ginza',
        108: 'Charmes',
        109: 'Stinson',
        111: 'Moon',
        112: 'Clarendon',
        113: 'Skyline',
        114: 'Gingham',
        115: 'Brooklyn',
        116: 'Ashby',
        117: 'Helena',
        118: 'Maven',
        603: 'Ludwig',
        605: 'Slumber',
        608: 'Perpetua',
        612: 'Aden',
        613: 'Juno',
        614: 'Reyes',
        615: 'Lark',
        616: 'Crema',
        640: 'BrightContrast',
        642: 'CrazyColor',
        643: 'SubtleColor',
    }

    @staticmethod
    def _get_closest_size(medias, width, height=0):
        """
        Try to extract a image/video object that will most match the resolution returned by the public API

        :param medias: list of images/videos
        :param width: desired width
        :param height: desired height
        :return:
        """
        current = None
        for media in medias:
            if not current:
                current = media
            if (abs(media['width'] - width) < abs(current['width'] - width) or
                    (media['width'] == current['width'] and not height and
                     not media['height'] == current['width']) or
                    (media['width'] == current['width'] and height and
                     abs(media['height'] - height) < abs(current['height'] - height))):
                current = media

        return current

    @staticmethod
    def _drop_keys(obj, keys):
        """
        Drop unwanted dict keys

        :param obj:
        :param keys:
        :return:
        """
        for k in keys:
            obj.pop(k, None)

    @classmethod
    def comment(cls, comment, drop_incompat_keys=False):
        """Patch a comment object"""
        comment['created_time'] = str(int(comment.get('created_at')))
        from_user = {
            'username': comment['user']['username'],
            'profile_picture': comment['user']['profile_pic_url'],
            'id': str(comment['user']['pk']),
            'full_name': comment['user']['full_name'],
        }
        comment['from'] = from_user
        comment['id'] = str(comment['pk'])
        if drop_incompat_keys:
            cls._drop_keys(
                comment,
                [
                    'bit_flags',
                    'content_type',
                    'created_at',
                    'created_at_utc',
                    'media_id',
                    'pk',
                    'status',
                    'type',
                    'user',
                    'user_id',
                ]
            )
        return comment

    @classmethod
    def media(cls, media, drop_incompat_keys=False):
        """Patch a media object"""
        media['link'] = 'https://www.instagram.com/p/{0!s}/'.format(media['code'])
        media['created_time'] = str(int(media.get('taken_at') or media.get('device_timestamp')))

        if media['media_type'] == MediaTypes.PHOTO:
            media['type'] = 'image'
        elif media['media_type'] == MediaTypes.VIDEO:
            media['type'] = 'video'
        elif media['media_type'] == MediaTypes.CAROUSEL:
            media['type'] = 'carousel'  # will be patched over below

        if media['caption']:
            media['caption']['id'] = str(media['caption']['pk'])
            media['caption']['created_time'] = str(int(media['caption']['created_at']))
            caption_from = {
                'username': media['caption']['user']['username'],
                'profile_picture': media['caption']['user']['profile_pic_url'],
                'id': str(media['caption']['user']['pk']),
                'full_name': media['caption']['user']['full_name'],
            }
            media['caption']['from'] = caption_from
            if drop_incompat_keys:
                cls._drop_keys(
                    media['caption'],
                    [
                        'bit_flags',
                        'content_type',
                        'created_at',
                        'created_at_utc',
                        'has_translation',
                        'media_id',
                        'pk',
                        'status',
                        'type',
                        'user',
                    ]
                )
        media['user'] = cls.list_user(media['user'], drop_incompat_keys=drop_incompat_keys)
        if media['media_type'] == MediaTypes.CAROUSEL and media.get('carousel_media', []):
            # patch carousel media
            for carousel_media in media.get('carousel_media', []):
                if carousel_media['media_type'] == MediaTypes.PHOTO:
                    carousel_media['type'] = 'image'
                elif carousel_media['media_type'] == MediaTypes.VIDEO:
                    carousel_media['type'] = 'video'
                image_versions2 = carousel_media.get('image_versions2', {}).get('candidates', [])
                images = {
                    'low_resolution': cls._get_closest_size(image_versions2, 320),
                    'thumbnail': cls._get_closest_size(image_versions2, 150, 150),
                    'standard_resolution': cls._get_closest_size(
                        image_versions2, carousel_media.get('original_width', 1000)),
                }
                carousel_media['images'] = images
                if carousel_media['media_type'] == MediaTypes.VIDEO:
                    video_versions = carousel_media.get('video_versions', [])
                    videos = {
                        'low_bandwidth': cls._get_closest_size(video_versions, 480),
                        'standard_resolution': cls._get_closest_size(
                            video_versions, carousel_media.get('original_width', 640)),
                        'low_resolution': cls._get_closest_size(video_versions, 640),
                    }
                    if drop_incompat_keys:
                        [cls._drop_keys(i, ['type']) for i in list(videos.values())]
                    carousel_media['videos'] = videos

                # patch user tags
                if carousel_media.get('usertags', {}).get('in', []):
                    usertags = carousel_media['usertags']['in']
                    user_tags = []
                    for ut in usertags:
                        pos = {'y': ut['position'][1], 'x': ut['position'][0]}
                        user = ut['user']
                        user['id'] = str(ut['user']['pk'])
                        user['profile_picture'] = ut['user']['profile_pic_url']
                        if drop_incompat_keys:
                            cls._drop_keys(user, ['profile_pic_url', 'pk', 'is_private'])

                        user_tags.append({
                            'position': pos,
                            'user': user,
                        })
                    carousel_media['users_in_photo'] = user_tags
                # patch location
                if 'location' not in carousel_media or not carousel_media['location'].get('lat'):
                    carousel_media['location'] = None
                else:
                    carousel_media['location']['latitude'] = carousel_media['location']['lat']
                    carousel_media['location']['longitude'] = carousel_media['location']['lng']
                    carousel_media['location']['id'] = carousel_media['location']['pk']

            first_carousel_media = media['carousel_media'][0]
            media['images'] = first_carousel_media['images']
            media['type'] = first_carousel_media['type']
            if first_carousel_media['media_type'] == MediaTypes.VIDEO:
                media['videos'] = first_carousel_media['videos']
        else:
            image_versions2 = media.get('image_versions2', {}).get('candidates', [])
            images = {
                'low_resolution': cls._get_closest_size(image_versions2, 320),
                'thumbnail': cls._get_closest_size(image_versions2, 150, 150),
                'standard_resolution': cls._get_closest_size(image_versions2, media.get('original_width', 1000)),
            }
            media['images'] = images

        if media['media_type'] == MediaTypes.VIDEO:
            video_versions = media.get('video_versions', [])
            videos = {
                'low_bandwidth': cls._get_closest_size(video_versions, 480),
                'standard_resolution': cls._get_closest_size(video_versions, media.get('original_width', 640)),
                'low_resolution': cls._get_closest_size(video_versions, 640),
            }
            if drop_incompat_keys:
                [cls._drop_keys(i, ['type']) for i in list(videos.values())]
            media['videos'] = videos

        likes = {
            'count': media.get('like_count', 0),
            'data': []
        }
        media['likes'] = likes
        comments = {
            'count': media.get('comment_count', 0),
            # Patch comment too
            'data': [
                cls.comment(c, drop_incompat_keys=drop_incompat_keys)
                for c in media.get('comments', [])
            ]
        }
        media['comments'] = comments
        if media.get('preview_comments'):
            [
                cls.comment(c, drop_incompat_keys=drop_incompat_keys)
                for c in media.get('preview_comments', [])
            ]

        media['attribution'] = None
        if media.get('filter_type') is not None and media.get('filter_type') in cls.FILTERS:
            media['filter'] = cls.FILTERS[media.get('filter_type')]
        else:
            media['filter'] = ''
        media['user_has_liked'] = media.get('has_liked', False)

        # Try to preserve location even if there's no lat/lng/pk
        if 'location' not in media or not media['location']:
            media['location'] = None
        elif (media.get('location', {}).get('lat')
              and media.get('location', {}).get('lng')
              and media.get('location', {}).get('pk')):
            media['location']['latitude'] = media['location']['lat']
            media['location']['longitude'] = media['location']['lng']
            media['location']['id'] = media['location']['pk']
        # For stories
        if (not media.get('location')
                and media.get('story_locations')
                and media.get('story_locations', [{}])[0].get('location')):
            story_location = media['story_locations'][0]['location']
            if (story_location.get('lat')
                    and story_location.get('lng')
                    and story_location.get('pk')):
                media['location'] = story_location

        media['tags'] = []
        if media.get('usertags', {}).get('in', []):
            usertags = media['usertags']['in']
            user_tags = []
            for ut in usertags:
                pos = {'y': ut['position'][1], 'x': ut['position'][0]}
                user = ut['user']
                user['id'] = str(ut['user']['pk'])
                user['profile_picture'] = ut['user']['profile_pic_url']
                if drop_incompat_keys:
                    cls._drop_keys(user, ['profile_pic_url', 'pk', 'is_private'])

                user_tags.append({
                    'position': pos,
                    'user': user,
                })
            media['users_in_photo'] = user_tags
        elif media.get('reel_mentions'):
            reel_mentions = media['reel_mentions']
            user_tags = []
            for rm in reel_mentions:
                pos = {'y': rm['y'], 'x': rm['x']}
                user = rm['user']
                user['id'] = str(rm['user']['pk'])
                user['profile_picture'] = rm['user']['profile_pic_url']
                if drop_incompat_keys:
                    cls._drop_keys(user, ['profile_pic_id', 'profile_pic_url', 'pk', 'is_private'])
                user_tags.append({
                    'position': pos,
                    'user': user,
                })
            media['users_in_photo'] = user_tags
        else:
            media['users_in_photo'] = []

        if drop_incompat_keys:
            cls._drop_keys(
                media,
                [
                    'can_viewer_save',
                    'caption_is_edited',
                    'client_cache_key',
                    'code',
                    'comment_count',
                    'comments_disabled',
                    'comment_likes_enabled',
                    'device_timestamp',
                    'filter_type',
                    'has_audio',
                    'has_liked',
                    'has_more_comments',
                    'image_versions2',
                    'is_reel_media',
                    'lat',
                    'like_count',
                    'lng',
                    'max_num_visible_preview_comments',
                    'media_type',
                    'next_max_id',
                    'organic_tracking_token',
                    'original_height',
                    'original_width',
                    'photo_of_you',
                    'pk',
                    'preview_comments',
                    'reel_mentions',
                    'saved_collection_ids',
                    'taken_at',
                    'top_likers',
                    'video_duration',
                    'video_versions',
                    'view_count',
                    'visibility',
                ]
            )
            if media['location']:
                cls._drop_keys(
                    media['location'],
                    [
                        'address',
                        'city',
                        'external_id',
                        'external_source',
                        'facebook_places_id',
                        'foursquare_v2_id',
                        'lat',
                        'lng',
                        'pk',
                        'state',
                    ]
                )
        return media

    @classmethod
    def user(cls, user, drop_incompat_keys=False):
        """Patch a user object """
        user['id'] = str(user['pk'])
        user['bio'] = user.get('biography', '')
        user['profile_picture'] = user['profile_pic_url']
        user['website'] = user.get('external_url', '')
        if 'media_count' in user and 'follower_count' in user and 'following_count' in user:
            counts = {
                'media': user['media_count'],
                'followed_by': user['follower_count'],
                'follows': user['following_count']
            }
            user['counts'] = counts
        if drop_incompat_keys:
            cls._drop_keys(
                user,
                [
                    'auto_expand_chaining',
                    'biography',
                    'external_lynx_url',
                    'external_url',
                    'follower_count',
                    'following_count',
                    'geo_media_count',
                    'has_anonymous_profile_picture',
                    'has_chaining',
                    'hd_profile_pic_url_info',
                    'hd_profile_pic_versions',
                    'include_direct_blacklist_status',
                    'is_business',
                    'is_favorite',
                    'is_private',
                    'is_unpublished',
                    'is_verified',
                    'media_count',
                    'pk',
                    'profile_context',
                    'profile_pic_id',
                    'profile_pic_url',
                    'usertags_count',
                ]
            )
        return user

    @classmethod
    def list_user(cls, user, drop_incompat_keys=False):
        """
        Patch a list user object, example in
        :meth:`Client.user_following`, :meth:`Client.user_followers`, :meth:`Client.search_users`
        """
        user['id'] = str(user['pk'])
        user['profile_picture'] = user['profile_pic_url']
        if drop_incompat_keys:
            cls._drop_keys(
                user,
                [
                    'byline',
                    'follower_count',
                    'friendship_status',
                    'has_anonymous_profile_picture',
                    'has_chaining',
                    'is_favorite',
                    'is_private',
                    'is_unpublished',
                    'is_verified',
                    'mutual_followers_count',
                    'pk',
                    'profile_pic_url',
                    'social_context',
                    'unseen_count',
                ]
            )
        return user



from .node import Node
from modeller import Model



attributes = lambda x: (x._name, x._data)

class Hashtag(Model, Node):
    _schema = {
        'properties':
            {
                'name': {'type': 'string'}
            },
        'type': 'object',
    }



from funcy import ignore, raiser, rcompose
from random import choice
from ..bot import Bot
from .common import decorate, substitute_vars, tap
from ..nodes import Comment, Media, User, Geotag




@decorate(accepts=Media, returns=Media)
def comment(bot: Bot, nodes,  args):

    try:
        max = float(args['max']) if 'max' in args else float('inf')
        comments = args['comments']
    except Exception:
        bot.logger.error('please add all necessary args, {} isn\'t enought'.format(args))
        return [], {}

    count = 0



    def increment():
        nonlocal count
        count += 1

    stop = raiser(StopIteration)



    return_if_suitable = lambda node: node \
        if bot.suitable(node, table='commented', specifier=str(comments)) \
        else tap(None,lambda: bot.logger.warning('{} not suitable'.format(node)))

    discard_if_reached_limit = lambda node: node \
        if not bot.reached_limit('comments') \
        else tap(None, bot.logger.error('reached commenting daily limit'))

    do_comment_from_groups = lambda node: map(
            lambda cmnts: do_comment(bot, choice(cmnts), node),
            comments) and node \
         if node else None



    process = rcompose(
        lambda x: stop() if x and count >= max else x,
        # return_if_suitable,
        # discard_if_reached_limit,
        do_comment_from_groups,
        lambda x: tap(x, increment) if x else None,
    )


    result = map(process, nodes)
    result = filter(lambda x: x, result)

    return result, {}




def do_comment(bot: Bot, text, node, thread_id=None):

    media_id = node.pk
    # print('caption', node.caption._yaml())
    evaluated_text = substitute_vars(text,
        author=ignore(Exception, '')(
            lambda: User(**node.user).username
        )(),
        # caption=node['caption']['text'],
        geotag=ignore(Exception, '')(
            lambda: Geotag(**node.location).name or ''
        )(),
        usertags=ignore(Exception, '')(
            lambda: list(map(lambda x: User(**x), node._usertags)) or []
        )()
    )

    bot.api.post_comment(
        media_id=media_id,
        comment_text=evaluated_text,
    )

    bot.logger.info('commented %s' % node)
    bot.total['comments'] += 1
    bot.sleep('comment')
    return node




class ClientDeprecationWarning(DeprecationWarning):
    pass


class ClientPendingDeprecationWarning(PendingDeprecationWarning):
    pass


class ClientExperimentalWarning(UserWarning):
    pass


class MediaTypes(object):
    """Psuedo enum-ish/lookup class for media types."""

    PHOTO = 1       #: Photo type
    VIDEO = 2       #: Video type
    CAROUSEL = 8    #: Carousel/Album type

    ALL = (PHOTO, VIDEO, CAROUSEL)

    __media_type_map = {
        'image': PHOTO,
        'video': VIDEO,
        'carousel': CAROUSEL,
    }

    @staticmethod
    def id_to_name(media_type_id):
        """Convert a media type ID to its name"""
        try:
            return [k for k, v in MediaTypes.__media_type_map.items() if v == media_type_id][0]
        except IndexError:
            raise ValueError('Invalid media ID')

    @staticmethod
    def name_to_id(media_type_name):
        """Convert a media type name to its ID"""
        try:
            return MediaTypes.__media_type_map[media_type_name]
        except KeyError:
            raise ValueError('Invalid media name')



from .user_feed import user_feed
from .hashtag_feed import hashtag_feed
from .geotag_feed import geotag_feed
from ..nodes import Geotag, Hashtag, Media, User
from itertools import chain
from .common import decorate


@decorate(accepts=(User, Hashtag, Geotag,), returns=Media)
def feed(bot, nodes, args):
    
    if isinstance(nodes, (list, tuple)):
        first = nodes[0]
    else:
        first = next(nodes, None)
        nodes = chain([first], nodes)

    switch = {
        isinstance(first, User): user_feed,
        isinstance(first, Hashtag): hashtag_feed,
        isinstance(first, Geotag): geotag_feed,
    }

    if True in switch:
        return switch[True](bot, nodes, args)

    else:
        return [], {}



import re

from ..compatpatch import ClientCompatPatch

USER_CHANNEL_ID_RE = r'^user_[1-9]\d+$'


class IGTVEndpointsMixin(object):
    """For endpoints in ``/igtv/``."""

    def tvchannel(self, channel_id, **kwargs):
        """
        Get channel

        :param channel_id: One of 'for_you', 'chrono_following', 'popular', 'continue_watching'
            (as returned by :meth:`tvguide`) or for a user 'user_12345' where user_id = '12345'
        """
        if (channel_id not in ('for_you', 'chrono_following', 'popular', 'continue_watching')
                and not re.match(USER_CHANNEL_ID_RE, channel_id)):
            raise ValueError('Invalid channel_id: {}'.format(channel_id))

        endpoint = 'igtv/channel/'
        params = {'id': channel_id}
        params.update(self.authenticated_params)
        if kwargs:
            params.update(kwargs)
        res = self._call_api(endpoint, params=params)

        if self.auto_patch:
            [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
             for m in res.get('items', [])]

        return res

    def tvguide(self):
        """TV guide to popular, following, suggested channels, etc"""
        res = self._call_api('igtv/tv_guide/')
        if self.auto_patch:
            for c in res.get('channels', []):
                [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
                 for m in c.get('items', [])]
            [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
             for m in res.get('my_channel', {}).get('items', [])]
        return res

    def search_igtv(self, text):
        """
        Search igtv

        :param text: Search term
        """
        text = text.strip()
        if not text.strip():
            raise ValueError('Search text cannot be empty')

        res = self._call_api('igtv/search/', query={'query': text})
        if self.auto_patch:
            for r in res.get('results', []):
                [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
                 for m in r.get('channel', {}).get('items', [])]
                if r.get('user'):
                    ClientCompatPatch.user(r['user'], drop_incompat_keys=self.drop_incompat_keys)
        return res



from typing import List
from funcy import mapcat, rcompose
from ..bot import Bot
from ..nodes import User
from .common import decorate, cycled_api_call


@decorate(accepts=User, returns=User)
def followers(bot: Bot, nodes,  args) -> List[User]:

    # bot.logger.debug('nodes at followers %s' % list(nodes)[:3])
    #
    # nodes = iter(list(nodes))
    amount = args.get('amount') or 1
    # query = args.get('query', {})

    pack_user = lambda item: User(**item)

    process = rcompose(
            lambda user: user.pk,
            lambda id: cycled_api_call(
                amount,
                bot,
                bot.api.user_followers,
                dict(user_id=id, **args.get('query', {}),) ,
                'users',
            ),
            lambda gen: map(pack_user, gen)

        )

    result = mapcat(process, nodes)

    return result, {}



from ..compatpatch import ClientCompatPatch


class UsertagsEndpointsMixin(object):
    """For endpoints in ``/usertags/``."""

    def usertag_feed(self, user_id, **kwargs):
        """
        Get a usertag feed

        :param user_id:
        :param kwargs:
        :return:
        """
        endpoint = 'usertags/{user_id!s}/feed/'.format(**{'user_id': user_id})
        query = {'rank_token': self.rank_token, 'ranked_content': 'true'}
        query.update(kwargs)
        res = self._call_api(endpoint, query=query)
        if self.auto_patch:
            [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
             for m in res.get('items', [])]
        return res

    def usertag_self_remove(self, media_id):
        """
        Remove your own user tag from a media post

        :param media_id: Media id
        :return:
        """
        endpoint = 'usertags/{media_id!s}/remove/'.format(**{'media_id': media_id})
        res = self._call_api(endpoint, params=self.authenticated_params)
        if self.auto_patch:
            ClientCompatPatch.media(res.get('media'))
        return res



from os import path
import os
from codecs import open
from setuptools import setup, find_packages
import os.path
here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, './README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Get the requirements
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    requirements = list(set(f.read().split('\n')) - set(['', '\n']))


# Get the requirements
with open(path.join(here, 'VERSION'), encoding='utf-8') as f:
    version = f.read().strip()


setup(
    name='instabotnet',
    version=version,

    description='Instagram readable yaml rpc api for easy instagram scheduling and promotion',
    long_description=long_description,
    long_description_content_type='text/markdown',

    author='Tommaso De Rossi',
    author_email='daer.tommy@gmail.com',
    license='Apache Software License 2.0',

    url='https://github.com/rpc-botnets/instagram-botnet',
    keywords=['instagram', 'bot', 'api', 'botnet'],
    install_requires=[
        *requirements,
    ],
    package_data={'': ['*.yaml', '*.json', '*.yml']},
    include_package_data=True,
    classifiers=[
        # How mature is this project? Common values are
        # 'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Information Technology',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    packages=['instabotnet'],
)




import time
import hmac
import base64
import hashlib
from random import randint
import os
from datetime import datetime
import re


VALID_UUID_RE = r'^[a-f\d]{8}\-[a-f\d]{4}\-[a-f\d]{4}-[a-f\d]{4}-[a-f\d]{12}$'


def raise_if_invalid_rank_token(val, required=True):
    if required and not val:
        raise ValueError('rank_token is required')
    if not re.match(VALID_UUID_RE, val):
        raise ValueError('Invalid rank_token: {}'.format(val))


def gen_user_breadcrumb(size):
    """
    Used in comments posting.

    :param size:
    :return:
    """
    key = 'iN4$aGr0m'
    dt = int(time.time() * 1000)

    # typing time elapsed
    time_elapsed = randint(500, 1500) + size * randint(500, 1500)

    text_change_event_count = max(1, size / randint(3, 5))

    data = '{size!s} {elapsed!s} {count!s} {dt!s}'.format(**{
        'size': size, 'elapsed': time_elapsed, 'count': text_change_event_count, 'dt': dt
    })
    return '{0!s}\n{1!s}\n'.format(
        base64.b64encode(hmac.new(key.encode('ascii'), data.encode('ascii'), digestmod=hashlib.sha256).digest()),
        base64.b64encode(data.encode('ascii')))


class Chunk(object):
    """
    Simple object class to encapulate an upload Chunk
    """
    def __init__(self, index, start, end, total):
        self.index = index
        self.start = start
        self.end = end
        self.total = total

    @property
    def is_first(self):
        return self.index == 0

    @property
    def is_last(self):
        return self.index == self.total - 1

    @property
    def length(self):
        return self.end - self.start


def get_file_size(fp):
    """
    Get the file size for a file-like object

    :param fp: file-like object
    :return:
    """
    original_file_position = fp.tell()
    fp.seek(0, os.SEEK_END)
    file_len = fp.tell()
    fp.seek(original_file_position, os.SEEK_SET)
    return file_len


def chunk_generator(chunk_count, chunk_size, file_data):
    """
    Generic chunk generator logic

    :param chunk_count: Number of chunks wanted
    :param chunk_size: Size of each chunk
    :param file_data: bytes to be split into chunk
    :return:
    """
    try:
        total_len = len(file_data)
        is_fp = False
    except TypeError:
        total_len = get_file_size(file_data)
        is_fp = True
    for i in range(chunk_count):
        start_range = i * chunk_size
        end_range = (start_range + chunk_size) if i < (chunk_count - 1) else total_len
        chunk_info = Chunk(i, start_range, end_range, chunk_count)
        if is_fp:
            file_data.seek(chunk_info.start, os.SEEK_SET)
            yield chunk_info, file_data.read(chunk_info.length)
        else:
            yield chunk_info, file_data[chunk_info.start: chunk_info.end]


def max_chunk_size_generator(chunk_size, file_data):
    """
    Generate chunks by defining a maximum chunk size

    :param chunk_size: Maximum chunk size allow
    :param file_data: bytes data
    :return:
    """
    try:
        file_len = len(file_data)
    except TypeError:
        file_len = get_file_size(file_data)
    chunk_count, final_chunk = divmod(file_len, chunk_size)
    if final_chunk:
        chunk_count += 1
    return chunk_generator(chunk_count, chunk_size, file_data)


def max_chunk_count_generator(chunk_count, file_data):
    """
    Generate chunks by defining a maximum number of chunks

    :param chunk_count: Max number of chunks
    :param file_data: bytes data
    :return:
    """
    try:
        # bytes data
        chunk_size = len(file_data) // chunk_count
    except TypeError:
        # file like object
        file_len = get_file_size(file_data)
        chunk_size = file_len // chunk_count

    return chunk_generator(chunk_count, chunk_size, file_data)


def ig_chunk_generator(file_data, max_chunk_size=(500 * 1024)):
    """
    Generate chunks in similar pattern to IG

    :param file_data: bytes data
    :param max_chunk_size:
    :return:
    """
    try:
        total_len = len(file_data)
        is_fp = False
    except TypeError:
        total_len = get_file_size(file_data)
        is_fp = True

    first_chunk_size = 200000
    max_chunk_size = max_chunk_size or (500 * 1024)
    chunks_generated = []
    last_yield_dt = None

    while sum(chunks_generated) < total_len:
        if not chunks_generated or total_len <= first_chunk_size:
            # first chunk
            chunk = Chunk(0, 0, min(first_chunk_size, total_len), 0)
            chunks_generated.append(chunk.length)
        else:
            chunk_elapsed_time = datetime.now() - last_yield_dt
            try:
                new_chunk_size = 5000 * chunks_generated[-1] / int(chunk_elapsed_time.total_seconds() * 1000)
            except ZeroDivisionError:
                new_chunk_size = max_chunk_size
            new_chunk_size = min(max_chunk_size, new_chunk_size)
            chunk_start = sum(chunks_generated)
            chunk_end = min(chunk_start + new_chunk_size, total_len)
            chunk = Chunk(
                len(chunks_generated), sum(chunks_generated), chunk_end,
                0 if chunk_end < total_len else len(chunks_generated) + 1)
            chunks_generated.append(chunk.length)

        last_yield_dt = datetime.now()
        if is_fp:
            file_data.seek(chunk.start, os.SEEK_SET)
            yield chunk, file_data.read(chunk.length)
        else:
            yield chunk, file_data[chunk.start: chunk.end]


class InstagramID(object):
    """
    Utility class to convert between IG's internal numeric ID and the shortcode used in weblinks.
    Does NOT apply to private accounts.
    """
    ENCODING_CHARS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_'

    @staticmethod
    def _encode(num, alphabet=ENCODING_CHARS):
        """Covert a numeric value to a shortcode."""
        if num == 0:
            return alphabet[0]
        arr = []
        base = len(alphabet)
        while num:
            rem = num % base
            num //= base
            arr.append(alphabet[rem])
        arr.reverse()
        return ''.join(arr)

    @staticmethod
    def _decode(shortcode, alphabet=ENCODING_CHARS):
        """Covert a shortcode to a numeric value."""
        base = len(alphabet)
        strlen = len(shortcode)
        num = 0
        idx = 0
        for char in shortcode:
            power = (strlen - (idx + 1))
            num += alphabet.index(char) * (base ** power)
            idx += 1
        return num

    @classmethod
    def weblink_from_media_id(cls, media_id):
        """
        Returns the web link for the media id

        :param media_id:
        :return:
        """
        return 'https://www.instagram.com/p/{0!s}/'.format(cls.shorten_media_id(media_id))

    @classmethod
    def shorten_media_id(cls, media_id):
        """
        Returns the shortcode for a media id

        :param media_id: string in the format id format: AAA_BB where AAA is the pk, BB is user_id
        :return:
        """
        # media id format: AAA_BB where AAA is the pk, BB is user_id
        internal_id = int((str(media_id).split('_')[0]))
        return cls.shorten_id(internal_id)

    @classmethod
    def shorten_id(cls, internal_id):
        """
        Returns the shortcode for a numeric media PK

        :param internal_id: numeric ID value
        :return:
        """
        return cls._encode(internal_id)

    @classmethod
    def expand_code(cls, short_code):
        """
        Returns the numeric ID for a shortcode

        :param short_code:
        :return:
        """
        return cls._decode(short_code)



import warnings

from .common import ClientExperimentalWarning, ClientDeprecationWarning
from ..compatpatch import ClientCompatPatch


class UsersEndpointsMixin(object):
    """For endpoints in ``/users/``."""

    def user_info(self, user_id):
        """
        Get user info for a specified user id

        :param user_id:
        :return:
        """
        res = self._call_api('users/{user_id!s}/info/'.format(**{'user_id': user_id}))
        if self.auto_patch:
            ClientCompatPatch.user(res['user'], drop_incompat_keys=self.drop_incompat_keys)
        return res

    def username_info(self, user_name):
        """
        Get user info for a specified user name
        :param user_name:
        :return:
        """
        res = self._call_api('users/{user_name!s}/usernameinfo/'.format(**{'user_name': user_name}))
        if self.auto_patch:
            ClientCompatPatch.user(res['user'], drop_incompat_keys=self.drop_incompat_keys)
        return res

    def user_detail_info(self, user_id, **kwargs):
        """
        EXPERIMENTAL ENDPOINT, INADVISABLE TO USE.
        Get user detailed info

        :param user_id:
        :param kwargs:
            - **max_id**: For pagination
            - **min_timestamp**: For pagination
        :return:
        """
        warnings.warn('This endpoint is experimental. Do not use.', ClientExperimentalWarning)

        endpoint = 'users/{user_id!s}/full_detail_info/'.format(**{'user_id': user_id})
        res = self._call_api(endpoint, query=kwargs)
        if self.auto_patch:
            ClientCompatPatch.user(res['user_detail']['user'], drop_incompat_keys=self.drop_incompat_keys)
            [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
             for m in res.get('feed', {}).get('items', [])]
            [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
             for m in res.get('reel_feed', {}).get('items', [])]
            [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
             for m in res.get('user_story', {}).get('reel', {}).get('items', [])]
        return res

    def user_map(self, user_id):    # pragma: no cover
        """
        Get a list of geo-tagged media from a user

        :param user_id: User id
        :return:
        """
        warnings.warn(
            'This endpoint is believed to be obsolete. Do not use.',
            ClientDeprecationWarning)

        endpoint = 'maps/user/{user_id!s}/'.format(**{'user_id': user_id})
        return self._call_api(endpoint)

    def search_users(self, query, **kwargs):
        """
        Search users

        :param query: Search string
        :return:
        """
        query_params = {
            'q': query,
            'timezone_offset': self.timezone_offset,
            'count': 50,
        }
        query_params.update(kwargs)
        res = self._call_api('users/search/', query=query_params)
        if self.auto_patch:
            [ClientCompatPatch.list_user(u, drop_incompat_keys=self.drop_incompat_keys)
             for u in res.get('users', [])]
        return res

    def check_username(self, username):
        """
        Check username

        :param username:
        :return:
            .. code-block:: javascript

                {
                  "status": "ok",
                  "available": false,
                  "username": "xxx",
                  "error_type": "username_is_taken",
                  "error": "The username xxx is not available."
                }
        """
        params = {'username': username}
        return self._call_api('users/check_username/', params=params)

    def blocked_user_list(self):
        """
        Get list of blocked users

        """
        return self._call_api('users/blocked_list/')

    def user_reel_settings(self):
        """
        Get user reel settings
        """
        res = self._call_api('users/reel_settings/')
        if self.auto_patch and res.get('blocked_reels', {}).get('users'):
            [ClientCompatPatch.list_user(u, drop_incompat_keys=self.drop_incompat_keys)
             for u in res.get('blocked_reels', {}).get('users', [])]
        return res

    def set_reel_settings(
            self, message_prefs,
            allow_story_reshare=None, reel_auto_archive=None,
            save_to_camera_roll=None):
        """
        Set story message replies settings

        :param message_prefs: One of 'anyone', 'following', 'off'
        :param allow_story_reshare: bool
        :param auto_archive: One of 'on', 'off'
        :param save_to_camera_roll: bool
        :return:
            .. code-block:: javascript

                {
                    "message_prefs": "off",
                    "status": "ok"
                }
        """
        if message_prefs not in ['anyone', 'following', 'off']:
            raise ValueError('Invalid message_prefs: {0!s}'.format(message_prefs))
        params = {'message_prefs': message_prefs}
        if allow_story_reshare is not None:
            params['allow_story_reshare'] = '1' if allow_story_reshare else '0'
        if reel_auto_archive is not None:
            if reel_auto_archive not in ['on', 'off']:
                raise ValueError('Invalid auto_archive: {0!s}'.format(reel_auto_archive))
            params['reel_auto_archive'] = reel_auto_archive
        if save_to_camera_roll is not None:
            params['save_to_camera_roll'] = '1' if save_to_camera_roll else '0'
        params.update(self.authenticated_params)
        return self._call_api('users/set_reel_settings/', params=params)




class Constants(object):
    """Constants holder class that stores the bulk of the fixed strings used in the library."""

    IG_SIG_KEY = '19ce5f445dbfd9d29c59dc2a78c616a7fc090a8e018b9267bc4240a30244c53b'
    IG_CAPABILITIES = '3brTvw=='
    SIG_KEY_VERSION = '4'
    APP_VERSION = '76.0.0.15.395'
    APPLICATION_ID = '567067343352427'
    FB_HTTP_ENGINE = 'Liger'

    ANDROID_VERSION = 24
    ANDROID_RELEASE = '7.0'
    PHONE_MANUFACTURER = 'samsung'
    PHONE_DEVICE = 'SM-G930F'
    PHONE_MODEL = 'herolte'
    PHONE_DPI = '640dpi'
    PHONE_RESOLUTION = '1440x2560'
    PHONE_CHIPSET = 'samsungexynos8890'
    VERSION_CODE = '138226743'

    USER_AGENT_FORMAT = \
        'Instagram {app_version} Android ({android_version:d}/{android_release}; ' \
        '{dpi}; {resolution}; {brand}; {device}; {model}; {chipset}; en_US; {version_code})'

    USER_AGENT_EXPRESSION = \
        r'Instagram\s(?P<app_version>[^\s]+)\sAndroid\s\((?P<android_version>[0-9]+)/(?P<android_release>[0-9\.]+);\s' \
        r'(?P<dpi>\d+dpi);\s(?P<resolution>\d+x\d+);\s(?P<manufacturer>[^;]+);\s(?P<device>[^;]+);\s' \
        r'(?P<model>[^;]+);\s(?P<chipset>[^;]+);\s[a-z]+_[A-Z]+;\s(?P<version_code>\d+)'

    USER_AGENT = USER_AGENT_FORMAT.format(**{
        'app_version': APP_VERSION,
        'android_version': ANDROID_VERSION,
        'android_release': ANDROID_RELEASE,
        'brand': PHONE_MANUFACTURER,
        'device': PHONE_DEVICE,
        'model': PHONE_MODEL,
        'dpi': PHONE_DPI,
        'resolution': PHONE_RESOLUTION,
        'chipset': PHONE_CHIPSET,
        'version_code': VERSION_CODE})

    LOGIN_EXPERIMENTS = 'ig_growth_android_profile_pic_prefill_with_fb_pic_2,ig_android_icon_perf2,ig_android_autosubmit_password_recovery_universe,ig_android_background_voice_phone_confirmation_prefilled_phone_number_only,ig_android_report_nux_completed_device,ig_account_recovery_via_whatsapp_universe,ig_android_stories_reels_tray_media_count_check,ig_android_background_voice_confirmation_block_argentinian_numbers,ig_android_device_verification_fb_signup,ig_android_reg_nux_headers_cleanup_universe,ig_android_reg_omnibox,ig_android_background_voice_phone_confirmation,ig_android_gmail_autocomplete_account_over_one_tap,ig_android_phone_reg_redesign_universe,ig_android_skip_signup_from_one_tap_if_no_fb_sso,ig_android_reg_login_profile_photo_universe,ig_android_access_flow_prefill,ig_android_email_suggestions_universe,ig_android_contact_import_placement_universe,ig_android_ask_for_permissions_on_reg,ig_android_onboarding_skip_fb_connect,ig_account_identity_logged_out_signals_global_holdout_universe,ig_android_account_switch_infra_universe,ig_android_hide_fb_connect_for_signup,ig_restore_focus_on_reg_textbox_universe,ig_android_login_identifier_fuzzy_match,ig_android_suma_biz_account,ig_android_session_scoping_facebook_account,ig_android_security_intent_switchoff,ig_android_do_not_show_back_button_in_nux_user_list,ig_android_aymh_signal_collecting_kill_switch,ig_android_nux_add_email_device,ig_android_multi_tap_login_new,ig_android_persistent_duplicate_notif_checker,ig_android_login_safetynet,ig_android_fci_onboarding_friend_search,ig_android_editable_username_in_reg,ig_android_phone_auto_login_during_reg,ig_android_one_tap_fallback_auto_login,ig_android_device_detection_info_upload,ig_android_updated_copy_user_lookup_failed,ig_fb_invite_entry_points,ig_android_hsite_prefill_new_carrier,ig_android_gmail_oauth_in_reg,ig_two_fac_login_screen,ig_android_reg_modularization_universe,ig_android_passwordless_auth,ig_android_sim_info_upload,ig_android_universe_noticiation_channels,ig_android_realtime_manager_cleanup_universe,ig_android_analytics_accessibility_event,ig_android_direct_main_tab_universe,ig_android_email_one_tap_auto_login_during_reg,ig_android_prefill_full_name_from_fb,ig_android_directapp_camera_open_and_reset_universe,ig_challenge_kill_switch,ig_android_video_bug_report_universe,ig_account_recovery_with_code_android_universe,ig_prioritize_user_input_on_switch_to_signup,ig_android_modularized_nux_universe_device,ig_android_account_recovery_auto_login,ig_android_hide_typeahead_for_logged_users,ig_android_targeted_one_tap_upsell_universe,ig_android_caption_typeahead_fix_on_o_universe,ig_android_retry_create_account_universe,ig_android_crosshare_feed_post,ig_android_abandoned_reg_flow,ig_android_remember_password_at_login,ig_android_smartlock_hints_universe,ig_android_2fac_auto_fill_sms_universe,ig_type_ahead_recover_account,ig_android_onetaplogin_optimization,ig_android_family_apps_user_values_provider_universe,ig_android_smart_prefill_killswitch,ig_android_direct_inbox_account_switching,ig_android_exoplayer_settings,ig_android_bottom_sheet,ig_android_publisher_integration,ig_sem_resurrection_logging,ig_android_login_forgot_password_universe,ig_android_hindi,ig_android_hide_fb_flow_in_add_account_flow,ig_android_dialog_email_reg_error_universe,ig_android_low_priority_notifications_universe,ig_android_device_sms_retriever_plugin_universe,ig_android_device_verification_separate_endpoint'   # noqa

    EXPERIMENTS = 'ig_camera_android_badge_face_effects_universe,ig_android_dash_lazy_load_audio,ig_android_stories_landscape_mode,ig_android_whitehat_options_universe,ig_android_fb_profile_integration_fbnc_universe,ig_android_vc_create_thread_upon_call_initiation_universe,ig_android_stories_seen_state_swipe_forward_universe,ig_android_realtime_mqtt_logging,ig_branded_content_show_settings_universe,ig_android_stories_server_coverframe,ig_android_direct_mutation_manager_handler_thread_universe,ig_android_ad_async_ads_universe,ig_android_camera_arengine_shader_caching_universe,ig_android_live_audiomanager_leak,ig_feed_lockdown,ig_android_interactions_preview_comment_impression_universe,ig_android_enable_igrtc_module,ig_android_direct_vm_activity_sheet,ig_android_appstate_logger,ig_android_direct_breeze_sheet,ig_android_camera_new_post_smile_universe,ig_android_live_video_position_source_universe,mi_viewpoint_feed_switchover,ig_feed_ranking_report_issue,ig_camera_android_areffect_photo_capture_universe,ig_rtc_use_dtls_srtp,ig_android_video_prefetch_feed_fix,ig_android_direct_remove_in_composer_camera_button_animation_universe,ig_android_live_fault_tolerance_universe,ig_android_igtv_autoplay_on_prepare,ig_android_main_feed_new_posts_indicator_universe,ig_android_audience_control,ig_android_stories_gif_upload_fix,ig_android_one_tap_fbshare,ig_android_startup_thread_priority,ig_android_stories_question_sticker_new_formats_universe,ig_android_business_transaction_in_stories_consumer,ig_android_search_impression_logging,ig_android_rtl_api28_textlayout_crash_universe,ig_android_direct_thread_sidebar_send_states,ig_fbns_push,ig_face_effect_ranking,ig_android_direct_albums,ig_search_null_state_universe,ig_android_stories_music_sticker_default_variation,ig_android_direct_update_thread_metadata_universe,ig_android_codec_high_profile,ig_android_inline_appeal,ig_rti_inapp_notifications_universe,ig_promote_last_used_destination_universe,ig_android_vc_directapp_integration_universe,allow_publish_page_universe,ig_android_maintabfragment,ig_android_skip_get_fbupload_universe,ig_android_low_data_mode,ig_android_enable_zero_rating,ig_android_main_feed_refresh_style_universe,ig_android_reverse_audio,ig_background_prefetch,ig_android_request_verification_badge,ig_android_http_stack_experiment_2017,ig_direct_android_24h_visual_perf,ig_android_live_thread_delay_for_mute_universe,ig_android_fb_topsearch_sgp_fork_request,ig_hashtag_display_universe,ig_android_banyan_migration,ig_android_heap_uploads,ig_android_cookie_serialization_optimization_universe,ig_android_mqtt_cookie_auth_memcache_universe,ig_android_stories_feeback_message_composer_entry_point,ig_android_instacrash_detection,ig_explore_2018_h2_account_rec_deduplication_android,ig_android_photo_hashing,ig_android_increase_fd_limit,ig_android_log_failed_image_download_retries,ig_android_live_use_timestamp_normalizer,ig_android_direct_async_thread_store,ig_android_persistent_nux,ig_android_story_accidentally_click_investigation_universe,ig_android_live_capture_translucent_navigation_bar,ig_android_churned_find_friends_redirect_to_discover_people,ig_android_vc_capture_universe,ig_android_story_reactions,ig_android_video_playback_retry_time_threshold,ig_android_mi_netego_long_event_removal_universe,ig_android_global_scheduler_infra,ig_end_of_feed_ranking_universe,ig_android_live_emoji_easter_egg_universe,ig_stories_in_feed_unit_design_universe,ig_android_ads_manager_pause_resume_ads_universe,ig_android_show_welcome_card_self_post_universe,ig_android_hashtag_header_display,ig_android_delay_coldstart_logging,ig_android_explore_grid_icon_removal_universe,ig_android_hashtag_contextual_feed_follow_button,ig_internal_research_settings,ig_smb_ads_basket_of_values_universe,ig_android_shopping_pdp_universe,ig_android_anr,ig_close_friends_v4,ig_android_feed_seen_state_with_view_info,ig_android_direct_visual_previews_in_thread,ig_promote_budget_warning_view_universe,ig_android_vc_camera_zoom_universe,ig_promote_daily_budget_multiplier_universe,ig_android_interactions_direct_share_comment_universe,ig_camera_android_supported_capabilities_api_universe,ig_android_post_recs_hide_from_author_universe,ig_android_biz_conversion_editable_profile_review_universe,ig_android_ad_increase_story_adpreload_priority_universe,ig_android_cache_video_autoplay_checker,ig_android_photo_fbuploader_config,ig_android_ad_watchlead_universe,ig_android_live_viewer_single_tap_invite_universe,ig_android_cold_start_json_delivery_improvement,ig_stories_suggestions_for_small_tray_universe,ig_shopping_catalog_selection_done_button,ig_inventory_connections,ig_android_fb_profile_integration_universe,ig_android_stories_weblink_creation,ig_android_live_start_broadcast_optimized_universe,ig_android_netgo_cta,ig_android_histogram_reporter,ig_android_vc_universe,ig_fb_graph_differentiation_no_fb_data,ig_android_network_cancellation,ig_android_live_presence_universe,ig_android_search_normalization_recipients,ig_android_connect_owned_page_universe,ig_android_downloaded_image_decode_universe,ig_android_realtime_stories_fetching,ig_android_hashtag_following,ig_android_felix_release_all_players_on_pause,ig_android_low_data_mode_backup_1,ig_android_share_claim_page_universe,ig_direct_holdout_h2_2018,ig_android_reactive_feed_like_count,ig_android_redirect_to_web_on_oembed_fail_universe,ig_camera_android_facetracker_v12_universe,ig_android_biz_qp_suggest_page,ig_android_direct_mutation_manager_job_scheduler,ig_android_continuous_video_capture,ig_android_live_skin_smooth,ig_promote_net_promoter_score_universe,ig_android_qp_features,ig_android_reel_raven_video_segmented_upload_universe,ig_android_biz_new_choose_category,ig_android_rate_limit_mediafeedviewablehelper,ig_android_shopping_post_tagging_redesign,ig_android_invite_xout_universe,ig_android_direct_permanent_video_upload_length,ig_android_sso_use_trustedapp_universe,ig_mi_impression_mainfeed_switchover,ig_android_remove_follow_all_fb_list,ig_android_save_all,ig_android_vc_call_screen_universe,ig_android_vc_join_timeout_universe,felix_android_video_quality,ig_eof_demarcator_style_universe,ig_shopping_post_insights,ig_android_shopping_more_from_business,ig_android_igtv_feed_trailer,ig_android_skip_video_render,ig_android_highlight_stickers_universe,ig_android_gap_rule_enforcer_universe,ig_android_interactive_listview_during_refresh,ig_android_ffmpeg_muxer_write_retry_universe,ig_android_main_feed_carousels_universe,ig_android_post_recs_show_more_button_universe,ig_android_live_suggested_live_expansion,ig_android_direct_inbox_cache_inbox_row_qes_universe,ig_android_video_cover_frame_universe,ig_android_abr_settings,ig_android_direct_app_hide_recents_header_in_recipient_picker_universe,ig_android_disk_usage_logging_universe,ig_android_story_sharing_universe,ig_android_optic_camera_warmup,ig_android_video_refactor_logger,ig_promote_lotus_universe,ig_stories_engagement_team_holdout_universe,ig_android_stories_gallery_video_segmentation,ig_promote_review_screen_title_universe,ig_android_direct_replace_inbox_camera_with_stories_camera,ig_explore_2018_post_chaining_account_recs_dedupe_universe,ig_android_igtv_save,ig_android_direct_presence_indicator,ig_android_asset_picker_improvements,ig_android_react_native_universe_kill_switch,ig_android_fs_new_gallery,android_ig_live_blacklisting,ig_android_qp_kill_switch,ig_android_new_contact_invites_entry_points_universe,ig_android_optic_feature_testing,ig_android_ad_leadgen_single_screen_universe,ig_android_stories_highlights_fast_navigation_universe,ig_android_vc_add_users_universe,ig_android_react_native_email_sms_settings_universe,ig_android_sticker_search_explorations,ig_android_business_id_conversion_universe,ig_android_business_promote_refresh_fb_access_token_universe,ig_android_selfupdate_jobscheduler,ig_android_fb_url_universe,ig_camera_android_profile_ar_notification_universe,ig_android_story_viewer_linear_preloading_count,ig_live_holdout_h2_2018,ig_android_vc_missed_call_notification_action_call_back,ig_android_stories_tray_in_viewer,ig_android_betamap_universe,ig_android_feed_video_mute_button_position,instagram_aat,ig_login_activity,ig_video_experimental_encoding_consumption_universe,ig_android_stories_share_extension_video_segmentation,ig_camera_android_black_feed_sticker_fix_universe,ig_android_camera_post_smile_low_end_universe,ig_android_import_page_post_after_biz_conversion,ig_android_direct_inbox_rv_configuration_universe,ig_android_feed_upload_progress,ig_vc_h2_2018_holdout_universe,ig_camera_android_superzoom_icon_position_universe,ig_android_live_dash_latency_manager,instagram_interests_holdout,ig_android_user_detail_endpoint,ig_android_click_to_direct_story_reaction_universe,ig_android_shopping_sidecar_editing,ig_android_interactions_new_comment_like_pos_universe,ig_android_reel_tray_item_impression_logging_viewpoint,ig_android_gif_framerate_throttling,ig_android_shopping_checkout_mvp,ig_android_live_save_to_camera_roll_limit_by_screen_size_universe,ig_end_of_feed_universe,ig_android_live_use_all_preview_sizes,ig_promote_post_insights_entry_universe,ig_hero_player,ig_stories_music_themes,ig_android_video_ffmpeg_muxer_universe,ig_android_live_follow_from_comments_universe,ig_android_profile_phone_autoconfirm_universe,ig_android_inline_notifications_recommended_user,ig_android_live_ama_universe,ig_android_camera_use_gl_for_postcapture_type,ig_android_insights_media_hashtag_insight_universe,ig_account_recs_in_chaining,ig_android_igtv_whitelisted_for_web,ig_fb_cross_posting_sender_side_holdout,ig_android_felix_feed_badging_tooltip_universe,ig_camera_gallery_button_thumbnail_universe,ag_family_bridges_2018_h2_holdout,ig_android_arengine_separate_prepare,ig_android_direct_visual_history,ig_android_employee_options_override,ig_android_share_product_universe,ig_camera_android_ar_platform_universe,ig_android_nametag,ig_android_netego_scroll_perf,ig_fbns_preload_default,ig_android_cover_frame_blacklist,android_cameracore_ard_ig_integration,ig_android_use_iterative_box_blur,ig_android_direct_inbox_recyclerview_pool_size,ig_android_clear_inflight_image_request,ig_android_audio_ingestion_params,ig_android_native_logcat_interceptor,ig_android_stories_separate_overlay_creation,ig_android_enable_liger_preconnect_universe,ig_android_hacked_account_reporting,ig_android_high_res_gif_stickers,ig_android_direct_remove_permanent_reactions_bar,ig_android_vod_abr_universe,ig_payments_paypal,ig_android_hashtag_feed_tabbed,ig_android_vc_participants_grid_universe,ig_android_video_decoder_retry,ig_android_enable_main_feed_reel_tray_preloading,ig_android_camera_upsell_dialog,ig_account_identity_2018_h2_lockdown_phone_global_holdout,ig_android_one_tap_sharesheet_fb_extensions,ig_android_country_code_fix_universe,ig_android_optic_fast_preview_restart_listener,ig_android_inline_appeal_show_new_content,ig_android_show_su_in_other_users_follow_list,ig_android_fb_family_navigation_badging_user,ig_android_video_scrubber_thumbnail_universe,ig_lockdown_feed_caption_length_universe,ig_camera_android_optimizations_2018_h2_universe,ig_stories_music_sticker,ig_android_optic_disable_post_capture_preview_restart,ig_android_vc_minimized_viewer_universe,ig_android_share_others_post_reorder,ig_android_low_data_mode_backup_5,ig_school_community_v2_universe,ig_android_post_live_expanded_comments_view_universe,ig_android_story_ad_cta_context_universe,ig_android_save_auto_sharing_to_fb_option_on_server,ig_android_igtv_chaining,ig_android_profile_private_banner,ig_android_stories_video_prefetch_kb,ig_android_direct_stories_in_direct_inbox,android_cameracore_preview_frame_listener2_ig_universe,ig_android_live_stop_broadcast_on_404,ig_android_live_skip_live_encoder_pts_correction,ig_android_show_twitter_name_universe,ig_android_direct_new_message_ranking,ig_android_render_iframe_interval,ig_android_direct_allow_multiline_composition,ig_android_place_search_profile_image,live_with_request_to_join_button_universe,ig_story_camera_reverse_video_experiment,ig_android_file_descriptor_limit,ig_android_stories_tray_fast_scroll_universe,ig_android_story_ad_text_limitation_universe,ig_android_cameracore_ar_text_plugin_universe,ig_android_direct_audience_upgrade_in_thread_camera,ig_android_felix,ig_android_media_share_icon,ig_android_archive_features_holdout_universe,ig_share_to_story_toggle_include_shopping_product,ig_two_fac_totp_enable,ig_android_camera_universe,ig_android_insights_creative_tutorials_universe,ig_android_qp_slot_cooldown_enabled_universe,ig_android_photos_qpl,ig_android_video_call_finish_universe,ig_hashtag_following_holdout_universe,ig_android_facebook_global_state_sync_frequency_universe,ig_android_global_scheduler_direct,ig_android_unify_video_player,ig_android_webrtc_icerestart_universe,ig_android_scroll_stories_tray_to_front_when_stories_ready,ig_android_mi_holdout_h1_2019,ig_android_interactions_permalink_replace_single_media_universe,ig_android_ttcp_improvements,ig_android_live_comment_fetch_frequency_universe,ig_android_directapp_instagram_deeplinking,ig_android_direct_inbox_recyclerview,ig_shopping_viewer_share_action,ig_android_hashtag_row_preparer,ig_eof_caboose_universe,ig_android_optic_new_features_implementation,ig_android_optic_new_zoom_controller,ig_android_direct_log_badge_count_inconsistent,ig_android_qp_clash_management_enabled_v4_universe,ig_android_hide_button_for_invite_facebook_friends,ig_android_activity_feed_impression_logger,ig_android_visualcomposer_inapp_notification_universe,ig_android_direct_sticker_gifs_in_thread,ig_android_optic_surface_texture_cleanup,ig_android_live_align_by_2_universe,ig_android_mobile_boost_universe,ig_android_network_util_cache_info,ig_android_camera_new_early_show_smile_icon_universe,ig_android_ads_profile_cta_feed_universe,ig_android_viewpoint_netego_universe,ig_android_direct_remix_visual_messages,ig_android_camera_new_tray_behavior_universe,ig_android_auto_advance_su_unit_when_scrolled_off_screen,ig_android_business_ix_universe,ig_vp9_hd_blacklist,ig_android_new_one_tap_nux_universe,ig_feed_experience,ig_android_business_new_navigation_universe,ig_stories_injection_tool_enabled_universe,ig_android_direct_import_google_photos2,ig_android_stories_text_format_emphasis,ig_android_direct_app_invites,ig_android_promote_fbauth_universe,ig_android_video_resize_operation,ig_android_stories_loading_automatic_retry,ig_android_live_end_redirect_universe,ig_android_following_hashtags_tooltip,ig_direct_max_participants,ig_android_stories_whatsapp_share,ig_android_low_data_mode_backup_2,ig_android_bitmap_attribution_check,ig_android_contact_invites_nux_universe,ig_android_search_page_v2,ig_android_direct_share_story_to_facebook,ig_android_stories_music_overlay,ig_android_direct_null_state_activation_cards,ig_android_fbupload_sidecar_video_universe,ig_android_tagging_combined_indicator,ig_android_direct_app_thread_presence_header,ig_android_react_native_restart_after_error_universe,ig_android_camera_attribution_in_direct,ig_android_contact_point_upload_rate_limit_killswitch,ig_android_profile,ig_android_additional_contact_in_nux,ig_android_profile_activation_cards_expanded,ig_android_view_and_likes_cta_universe,ig_android_story_reactions_producer_holdout,ig_android_live_use_rtc_upload_universe,ig_android_live_replay_highlights_universe,ig_main_activity_cold_start,ig_android_direct_double_tap_like_everything,ig_android_direct_character_limit,ig_business_dynamic_conversion_universe,ig_android_shopping_channel_in_explore,ig_stories_holdout_h1_2018,ig_android_scroll_perf_qpl_killswitch,ig_android_fbns_optimization_universe,ig_camera_ar_effect_attribution_position,ig_android_video_ta_universe,ig_android_live_view_profile_from_comments_universe,ig_android_interactions_threaded_comments_in_feed_universe,ig_fbns_blocked,ig_android_sso_kototoro_app_universe,ig_android_stories_question_sticker_music_format,ig_android_biz_auto_slide_props,ig_media_account_rollout_universe,ig_android_show_fbunlink_button_based_on_server_data,ig_android_fs_creation_flow_tweaks,ig_android_recommend_accounts_killswitch,ig_android_shopping_save_product_collection_cell_redesign_universe,ig_android_direct_inbox_background_view_models,ig_android_page_claim_deeplink_qe,ig_android_profile_edit_phone_universe,ig_android_switch_back_option,ig_android_new_orders_entrypoint,ig_android_media_rows_async_inflate,ig_android_direct_story_chaining_v2,ig_android_ad_show_full_name_universe,ig_android_private_highlights_universe,ig_android_igtv_audio_always_on,ig_android_interactions_inline_composer_extensions_universe,ig_android_scroll_main_feed,ig_business_integrity_ipc_universe,ig_android_location_page_info_page_upsell,ig_camera_android_bg_processor,ig_android_stories_viewer_prefetch_improvements,ig_android_rate_limit_feed_item_viewable_helper,ig_android_fci_empty_feed_friend_search,ig_feed_requests_logs_universe,ig_android_video_qp_logger_universe,ig_nametag_data_collection,ig_discovery_holdout_universe,ig_android_recyclerview_binder_group_enabled_universe,ig_android_direct_create_shortcut,ig_android_ar_effects_button_display_timing,ig_vc_holdout_universe_h2,ig_android_stories_sampled_progress,ig_android_qpl_queue_time_universe,ig_android_downloadable_vp8_module,ig_android_ccu_jobscheduler_outer,ig_android_stories_viewer_modal_activity,ig_android_direct_thread_composer,ig_android_fbns_preload_direct_universe,ig_android_direct_24h_replayability_nux_killswitch_universe,ig_android_activity_feed_row_click,ig_android_time_spent_dashboard,ig_android_loom_v2,ig_android_ad_pbia_header_click_universe,ig_android_direct_quick_replies,ig_android_handle_username_in_media_urls_universe,ig_android_request_compression_universe,ig_android_live_pip_minimize_universe,ig_android_usersession_leak_patching_universe,ig_android_stories_viewer_tall_android_cap_media_universe,ig_android_growth_fci_team_holdout_universe,ig_android_insights_holdout,ig_feed_engagement_holdout_2018_h1,ig_fb_graph_differentiation_only_fb_candidates,ig_pacing_overriding_universe,ig_android_direct_app_multi_account_badging,ig_android_direct_persisted_text_drafts_universe,ig_android_felix_prefetch_thumbnail_sprite_sheet,ig_camera_android_segmentation_async_universe,ig_android_category_search_in_sign_up,ig_android_separate_network_executor,ig_android_interactions_comment_like_for_all_feed_universe,ig_android_remove_push_notifications,ig_android_video_segment_ffmpeg_muxer_universe,ig_android_downgrade_viewport_exit_behavior,ig_android_vc_call_ended_cleanup_universe,ig_android_universe_video_production,ig_android_intialization_chunk_410,ig_android_live_analytics,ig_android_stories_music_filters,ig_android_camera_gallery_upload_we_universe,ig_android_video_exoplayer_2,ig_android_stories_music_precapture,ig_android_bitmap_compress_retry_universe,ig_android_verified_comments_universe,ig_android_dash_script,ig_android_igtv_feed_banner_redesign,ig_shopping_viewer_intent_actions,ig_android_gallery_order_by_date_taken,ig_android_location_plugin_leak_detection,ig_android_custom_story_import_intent,ig_lockdown_feed_perf,ig_android_camera_ar_platform_profile_universe,ig_stories_allow_camera_actions_while_recording,ig_android_optic_new_architecture,ig_android_ig_to_fb_sync_universe,ig_android_fbc_upsell_on_dp_first_load,ig_android_video_watermark_universe_qe2,ig_android_shopping_video_product_tag_consumption,ig_android_share_others_post_share_sheet,ig_biz_growth_entry_value,ig_android_stories_alignment_guides_universe,ig_android_livewith_guest_adaptive_camera_universe,ig_android_business_transaction_in_stories_creator,ig_android_optic_thread_priorities,ig_android_delayed_comments,ig_profile_company_holdout_h2_2018,ig_android_feed_coldstart_universe,ig_android_felix_pager_center_buffer_bias,ig_android_edit_metadata,ig_android_user_url_deeplink_fbpage_endpoint,ig_android_direct_face_filter_button_in_composer,ig_android_stories_helium_balloon_badging_universe,ig_android_rate_limit_feed_video_module,ig_android_ad_watchbrowse_universe,ig_android_stories_private_mention_sharing_universe,ig_direct_raven_search_universe,ig_android_live_pivot_to_reshare_universe,ig_company_profile_holdout,ig_android_invite_list_button_redesign_universe,ig_android_log_mediacodec_info,ig_android_fb_follow_server_linkage_universe,ig_android_direct_expiring_media_loading_errors,ig_android_direct_remove_blurred_profile_photo_for_thread_camera_universe,ig_camera_regiontracking_use_similarity_tracker_for_scaling,ig_android_not_modified_cache_universe,ig_android_direct_thread_green_dot_presence_universe,ig_igds_snackbar_android_universe,ig_android_insights_relay_optimization_universe,ig_android_stories_viewer_bitmap_holder,ig_android_shopping_catalogsearch,ig_android_location_page_intent_survey,ig_android_reel_zoom_universe,ig_android_biz_suggested_category,ig_android_cpu_frame_rendering_universe,ig_android_stories_create_flow_favorites_tooltip,ig_android_q3lc_transparency_control_settings,ig_android_stories_music_broadcast_receiver,ig_android_direct_send_new_combined_reshare,ig_android_resuming_failed_image_downloads_universe,ig_android_push_notifications_settings_redesign_universe,ig_android_enable_request_compression_ccu,ig_android_vc_ongoing_call_notification_universe,ig_android_stories_helium_long_press_universe,ig_fb_notification_universe,ig_branded_content_paid_branded_content,ig_android_downloadable_igrtc_module,ig_android_hide_reset_with_fb_universe,ig_android_direct_newer_single_line_composer_universe,ig_android_story_decor_image_fbupload_universe,ig_android_hashtag_creation_development,ig_android_ad_view_ads_native_universe,ig_android_hero_player_settings,ig_promote_ppe_v2_universe,ig_android_stories_archive_calendar,ig_android_ad_watchbrowse_cta_universe,ig_android_player_crash_report,ig_business_signup_biz_id_universe,ig_android_video_render_device_tiers,ig_android_payload_based_scheduling,ig_android_realtime_iris,ig_android_direct_gifs_in_thread,ig_android_main_feed_fragment_scroll_timing_histogram_uni,ig_android_direct_inbox_recyclerview_fixedsize_universe,ig_android_qp_batch_fetch_caching_enabled_v1_universe,ig_android_inline_editing_local_prefill,ig_android_location_feed_related_business,ig_promote_audience_selection_universe,ig_android_direct_low_contrast_inbox,ig_android_media_rows_prepare_10_31,ig_android_stories_fix_current_active_item_bound_crash,ig_family_bridges_holdout_universe,ig_android_push_notification_settings_universe,ig_android_updatelistview_on_loadmore,ig_promote_no_create_ads_check_universe,ig_android_business_ix_self_serve,ig_direct_raven_sharesheet_ranking,ig_android_insta_video_consumption_infra,ig_android_api_urlencode_universe,ig_android_concurrent_cold_start_universe,ig_android_direct_inbox_custom_rv_prefetch,ig_android_vc_missed_call_notification_action_reply,ig_android_multi_capture_camera,ig_android_stories_cross_sharing_to_fb_holdout_universe,ig_smb_ads_holdout_2018_h2_universe,instagram_android_stories_sticker_tray_redesign,ig_android_edit_location_page_info,ig_android_felix_video_upload_length,ig_android_video_segment_resume_policy_universe,ig_android_igsystrace_universe,ig_android_direct_split_reshares,ig_android_igtv_banner_changes,ig_android_dash_for_vod_universe,ig_android_new_highlight_button_text,ig_android_video_call_participant_state_caller_universe,ig_android_story_ads_default_long_video_duration,ig_android_stories_camera_enhancements,ig_android_feed_stale_check_interval,ig_find_loaded_classes,ig_android_interactions_realtime_typing_indicator_and_live_comments,ig_android_video_live_trace_universe,ig_android_stories_gallery_improvements,ig_close_friends_v4_global,ig_android_stories_large_reel_navigation,ig_android_prefetch_notification_data,ig_android_3pspp,ig_android_direct_new_intro_card,ig_android_direct_pending_media,ig_camera_ar_image_transform_library,ig_android_live_share_post_live_universe,ig_android_comments_composer_newline_universe,ig_android_direct_mutation_manager_iris,ig_android_stories_gif_sticker,ig_android_interactions_feed_dwell_universe,ig_camera_android_superzoomv3_attribution_universe,ig_android_stories_posting_offline_ui,ig_camera_android_superzoomv3_universe,ig_android_account_hierarchy_account_association_signal_upload_kill_switch,ig_android_offline_mode_holdout,ig_android_comments_direct_reply_to_author,ig_android_video_streaming_upload_universe,ig_direct_holdout_h1_2019,ig_android_stepper_header,ig_android_family_bridge_discover,ig_direct_report_conversation_universe,igds_android_listrow_migration_universe,ig_android_camera_sdk_check_gl_surface_r2,ig_promote_story_insights_entry_universe,ig_android_http_service_same_thread,ig_challenge_general_v2,ig_android_expired_build_lockout,ig_android_felix_keep_video_view,ig_feed_video_autoplay_tap_threshold,ig_android_vpvd_impressions_universe,ig_android_stories_reel_interactive_tap_target_size,ig_android_rendering_controls,ig_android_os_version_blocking,ig_promote_fix_expired_fb_accesstoken_android_universe,ig_android_stories_combined_asset_search,ig_android_interactions_emoji_extension_followup_universe,ig_android_shopping_native_catalog_selection,ig_android_profile_unified_follow_view,ig_android_igtv_no_badge,ig_android_unfollow_from_main_feed_v2,ig_android_livewith_liveswap_optimization_universe,ig_promote_video_retry_universe,ig_android_vc_participant_state_callee_universe,ig_helium_v1,ig_android_buffered_analytics_logger_thread_safe,ig_android_fb_connect_follow_invite_flow,ig_android_video_stitch_after_segmenting_universe,ig_android_enable_swipe_to_dismiss_for_all_dialogs,ig_android_business_cross_post_with_biz_id_infra,ig_android_paid_branded_content_rendering,ig_android_rage_shake_whitelist,ig_android_low_data_mode_backup_4,ig_mi_analytics_uploader_diagnostics,ig_android_shopping_pdp_craft,ig_android_ad_connection_manager_universe,ig_android_skip_button_content_on_connect_fb_universe,ig_android_reset_to_feed_from_background,ig_android_ad_watchbrowse_carousel_universe,android_cameracore_ig_gl_oom_fixes_universe,ig_android_video_feed_universe,ig_android_hybrid_bitmap_version_2,ig_android_update_items_checks,ig_android_interactions_mention_search_presence_dot_universe,ig_android_direct_app_reel_grid_search,ig_android_live_disable_speed_test_ui_timeout_universe,ig_android_hashtag_page_reduced_related_items,ig_android_direct_mutation_manager_media_2,ig_direct_reshare_sharesheet_ranking,ig_android_image_fail_callback_fix_universe,ig_android_igtv_reshare,ig_direct_reshare_search_universe,ig_android_shopping_pdp_platformization,ig_branded_content_share_to_facebook,ig_android_building_aymf_universe,ig_android_stories_viewer_as_modal_high_end_launch,ig_android_collect_os_usage_events_universe,ig_android_shopping_product_appeals_universe,ig_android_direct_mqtt_send,ig_android_business_profile_share_link_universe,ig_android_reliability_leak_fixes_h2_2018,ig_promote_unified_insights_universe,ig_android_global_prefetch_scheduler,ig_fbns_shared,ig_android_stories_reel_media_item_automatic_retry,ig_android_interactions_composer_extensions_universe,ig_android_cache_timespan_objects,ig_android_rn_ads_manager_universe,ig_smb_ads_click_to_direct,ig_android_foreground_location_collection,ig_kill_connectivity_change_receiver,ig_android_pending_actions_serialization,ig_android_2018_h1_hashtag_report_universe,ig_android_new_camera_design_universe,ig_android_prefetch_carousels_on_swipe_universe,ig_android_ads_history_universe,ig_fb_graph_differentiation_top_k_fb_coefficients,ig_explore_2018_topic_channel_navigation_android_universe,ig_android_shopping_profile_tab_universe,ig_android_hashtag_unfollow_from_main_feed,ig_android_ad_watchmore_entry_point_universe,ig_android_stories_feedback_badging_universe,ig_android_low_latency_consumption_universe,ig_android_graphql_survey_new_proxy_universe,ig_android_resumable_downloads_logging_universe,ig_direct_recipients_search_universe,ig_android_scheduled_executor,ig_android_fblocation_universe,ig_promote_rename_to_boost_universe,ig_android_early_storyrequest,ig_android_ad_holdout_watchandmore_universe,ig_android_felix_insights,ig_android_interests_netego_dismiss,ig_android_realtime_always_start_connection_on_condition_universe,ig_android_split_contacts_list,ig_android_igtv_always_show_browse_ui,ig_android_always_use_server_recents,ig_android_carousel_prefetch_bumping,ig_fbns_kill_switch,ig_android_direct_send_thread_summary_fix_universe,ig_android_video_fix_logger,ig_stories_question_sticker_music_format_prompt,ig_mi_extra_bundle_investigation_universe,ig_camera_android_segmentation_qe2_universe,ig_android_direct_media_forwarding,ig_android_stories_close_friends_disable_first_time_badge,ig_android_reel_viewer_fetch_missing_reels_universe,ig_android_fb_link_ui_polish_universe,ig_android_signup_error_test,ig_android_video_webrtc_textureview,ig_android_business_promote_tooltip,mi_viewpoint_viewability_universe,ig_android_volume_controls,ig_xplat_shopping_cataloglist,ig_android_interactions_in_feed_comment_view_universe,ig_android_biz_category_prefill_universe,ig_android_pigeon_sampling,ig_android_gallery_high_quality_photo_thumbnails,ig_android_show_weekly_ci_upsell_limit,ig_android_tagging_video_preview,ig_direct_android_reply_modal_universe,ig_ei_option_setting_universe,ig_perf_android_holdout,ig_direct_core_holdout_q1_2018,ig_promote_insights_video_views_universe,ig_android_list_redesign,ig_android_claim_location_page,ig_android_search_normalization,ig_android_not_decoding_prefetch,ig_smb_review_screen_content_update_universe,ig_android_category_search_edit_profile,ig_android_direct_forward_messages_universe,ig_android_pbia_proxy_profile_universe,ig_android_cover_frame_rendering,ig_android_feed_post_sticker_alt,ig_camera_android_segmentation_enabled_universe,ig_android_shopping_profile_shop_redesign,ig_android_upload_retry_job_service,ig_android_stories_better_error_state_handling,ig_android_vc_in_app_notification_universe,ig_android_persistent_duplicate_notif_checker_user_based,ig_android_react_native_ota,ig_android_profile_memories_universe,ig_fb_graph_differentiation_control,ig_android_low_data_mode_backup_3,android_ig_camera_ar_asset_manager_improvements_universe,ig_android_explore_discover_people_entry_point_universe,ig_android_qcc_perf,ig_android_video_cache_evictor_universe,ig_android_limit_ashmem_cleanup_thread,ig_android_direct_business_holdout,ig_android_promote_feed_to_stories_universe,ig_media_geo_gating,ig_music_dash,ig_android_media_as_sticker,ig_android_internal_sticker_universe,ig_android_video_watermark_universe,ig_android_live_ama_viewer_universe,ig_android_live_streaming_experimental_abr_universe,ig_android_cronet_stack,ig_android_mention_sharing_from_reel_viewer_universe,ig_android_warm_headline_text,ig_android_new_block_flow,ig_android_story_landscape_ad_new_layout_universe,ig_android_long_form_video,ig_android_network_trace_migration,ig_android_story_ads_direct_cta_universe,ig_android_live_subscribe_user_level_universe,ig_android_ad_iab_qpl_kill_switch_universe,ig_android_fb_sync_options_universe,ig_android_saved_product_store,ig_android_stories_reappearing_tray_universe,ig_android_new_camera_design_container_animations_universe,ig_android_stories_disable_highlights_media_preloading,ig_fb_graph_differentiation,ig_android_logging_metric_universe_v2,ig_android_stories_persistent_tray_universe,ig_android_screen_recording_bugreport_universe,ig_android_friends_sticker,ig_android_whats_app_contact_invite_universe,ig_android_feed_auto_share_to_facebook_dialog,ig_android_felix_creation_enabled,ig_direct_android_larger_media_reshare_style,ig_android_stories_auto_retry_reels_media_and_segments,ig_android_image_mem_cache_strong_ref_universe,ig_direct_android_inbox_filter_for_all_universe,ig_android_suggested_highlights,ig_direct_giphy_gifs_rating,ig_stories_holdout_h2_2017,ig_android_fbpage_on_profile_side_tray,ig_android_video_server_coverframe,ig_android_video_controls_universe,ig_camera_holdout_h1_2018_performance,ig_android_stories_music_search_typeahead,ig_android_inappnotification_rootactivity_tweak,ig_android_local_info_page,ig_camera_holdout_h1_2018_product,ig_shopping_checkout_mvp_experiment,ig_android_hide_type_mode_camera_button,ig_timestamp_public_test,ig_android_webrtc_renderer_reuse_universe,ig_android_business_conversion_value_prop_v2,ig_android_live_wave_production_universe,ig_android_share_publish_page_universe,ig_android_question_sticker_replied_state,ig_android_early_feedrequest,ig_android_hashtag_search_suggestions,ig_android_hashtag_discover_tab,ig_android_leak_detector_upload_universe,ig_android_hashtag_page_support_places_tab,ig_android_cover_frame_retrieval,ig_android_live_bg_download_face_filter_assets_universe,ig_android_direct_continuous_capture,ig_android_search_hashtag_badges,ig_android_direct_tabbed_media_picker,ig_android_video_ssim_report_universe,ig_android_direct_view_more_qe,ig_camera_android_effect_info_bottom_sheet_universe,ig_promote_add_payment_navigation_universe,ig_android_direct_voice_messaging,ig_android_signup_refactor_santity,ig_android_profile_lazy_load_carousel_media,ig_android_reel_dashboard_camera_entry_point,ig_android_su_follow_back,ig_android_direct_reel_options_entry_point_2_universe,ig_android_ad_redesign_iab_universe,ig_android_universe_reel_video_production,ig_android_power_metrics,ig_android_modal_activity_no_animation_fix_universe,ig_android_bitmap_cache_executor_size,ig_android_direct_log_badge_count,ig_android_direct_remove_visual_messages_nuxs,ig_android_creation_new_post_title,ig_camera_fast_tti_universe,ig_android_non_square_first,ig_promote_media_picker_universe,ig_android_direct_thread_content_picker,ig_android_vc_fix_joining_other_call_with_new_intent,ig_android_drawable_usage_logging_universe,ig_android_reel_viewer_data_buffer_size,ig_android_hashtag_contextual_feed_account_recs,ig_traffic_routing_universe,ig_promote_political_ads_universe,ig_android_clarify_invite_options,ig_android_igtv_aspect_ratio_limits,ig_android_effect_tray_background,ig_android_disable_scroll_listeners,ig_android_profile_neue_universe,ig_android_create_page_on_top_universe,ig_stories_selfie_sticker,ig_android_video_upload_quality_qe1,ig_android_mobile_http_flow_sampling_weight_universe,ig_android_stories_music_awareness_universe,ig_android_live_nerd_stats_universe,ig_android_video_cache_size_universe,ig_camera_android_focus_attribution_universe,ig_android_promote_story_to_story_universe,ig_android_igds_edit_profile_fields,ig_android_reel_impresssion_cache_key_qe_universe,ig_video_holdout_h2_2017,ig_android_immersive_viewer_follow,ig_android_sso_family_key_universe,ig_android_direct_share_sheet_custom_fast_scroller,ig_android_external_gallery_import_affordance,ufi_share,ig_android_sonar_prober_universe,ig_android_swipe_up_area_universe,ig_android_video_segmented_upload_universe,ig_perf_android_holdout_2018_h1,ig_android_live_special_codec_size_list,ig_android_view_info_universe,ig_android_shopping_combined_tagging_universe,ig_android_cold_start_cool_off_universe,ig_android_shopping_video_product_tag_creation,ig_android_startup_sampling_rate_universe,ig_android_igtv_new_browse,ig_android_story_import_intent,ig_android_direct_inbox_typing_indicator,ig_android_edit_highlight_redesign,ig_android_insta_video_broadcaster_infra_perf,ig_android_live_webrtc_livewith_params,ig_android_show_fb_name_universe,ig_android_fix_prepare_direct_push,ig_android_stories_viewer_responsiveness_universe,ig_android_interactions_show_verified_badge_for_preview_comments_universe,ig_android_stories_skip_preload_to_launch_viewer,ig_android_live_start_live_button_universe,ig_android_direct_speed_cam_univ,ig_android_profile_menu_reorder_universe,ig_android_acra_double_oom_reservation,ig_android_live_viewer_tap_to_hide_chrome_universe,ig_android_vc_sounds_universe,ig_android_igtv_native_pip,ig_android_igtv_refresh_tv_guide_interval,ig_direct_inbox_search_universe,ig_android_experimental_onetap_dialogs_universe,ig_android_pendingmedia_retry,ig_android_settings_redesign,ig_android_direct_search_story_recipients_universe,ig_android_fb_sharing_shortcut,ig_android_direct_segmented_video,ig_android_grid_cell_count,ig_android_ad_watchinstall_universe,ig_android_realtime_manager_optimization,ig_android_shortcuts,ig_android_comments_notifications_universe,ig_android_vc_webrtc_params,ig_android_critical_path_manager_universe,ig_android_canvas_tilt_to_pan_universe,ig_android_feed_sharing_memory_leak,ig_android_ad_account_top_followers_universe,ig_android_offline_reel_feed,ig_promote_review_screen_universe,ig_android_vc_end_screen_user_feedback_universe,ig_android_vc_use_timestamp_normalizer,native_contact_invites_universe,ig_android_feed_post_sticker,ig_android_facebook_crosspost,ig_android_local_2018_h2_holdout,ig_android_stories_tray_refresh_universe,ig_android_viewer_tapback_size_universe,ig_android_nametag_save_experiment_universe,ig_promote_estimated_clicks_universe,ig_business_profile_18h1_holdout_universe,ig_android_nearby_venues_location_timeout_fallback,ig_android_category_clickable_rows_ui,ig_android_photo_invites,ig_interactions_h2_2018_team_holdout_universe,ig_branded_content_tagging_upsell,ig_android_ccu_jobscheduler_inner,ig_android_story_ads_instant_sub_impression_universe,ig_explore_2018_finite_chain_android_universe,ig_android_gqls_typing_indicator,ig_android_direct_visual_message_prefetch_count_universe,ig_android_webrtc_encoder_factory_universe,ig_ads_increase_connection_step2_v2,ig_scroll_by_two_cards_for_suggested_invite_universe,ig_android_internal_collab_save'  # noqa




from typing import List
from funcy import  rcompose, mapcat
from itertools import islice
from ..bot import Bot
from ..nodes import  Media, Geotag
from .common import cycled_api_call, decorate, tap




@decorate(accepts=Geotag, returns=Media)
def geotag_feed(bot: Bot, nodes,  args) -> List[Media]:

    pack_media = lambda data: Media(**data['media'])
    amount = args.get('amount') or 1
    ordering = args.get('ordering', 'ranked')

    process = rcompose(
        lambda tag: tag.id,
        # lambda x: tap(x, lambda: print(x)),
        lambda id: cycled_api_call(None, bot, bot.api.location_section, dict(location_id=id, tab=ordering), ( 'sections',)),
        # lambda x: tap(x, lambda: print(next(x))),
        lambda gen: mapcat(lambda data: data['layout_content']['medias'], gen),
        lambda gen: islice(gen, amount),
        lambda items: map(pack_media, items),
    )

    result = mapcat(process, nodes)

    return result, {}




from typing import List
from funcy import  rcompose, mapcat
from itertools import islice
from ..bot import Bot
from ..nodes import  Story, Geotag
from .common import cycled_api_call, decorate, tap




@decorate(accepts=Geotag, returns=Story)
def geotag_stories(bot: Bot, nodes,  args) -> List[Story]:

    pack_story = lambda data: Story(**data)
    amount = args.get('amount') or 1

    process = rcompose(
        lambda tag: tag.id,
        # lambda x: tap(x, lambda: print(x)),
        lambda id: cycled_api_call(amount, bot, bot.api.location_stories, id, ( 'story', 'items')),
        # lambda x: tap(x, lambda: print(next(x))),
        # lambda gen: islice(gen, amount),
        lambda items: map(pack_story, items),
    )

    result = mapcat(process, nodes)

    return result, {}



import json
import re
import warnings
import time
from random import randint

from .common import ClientExperimentalWarning, MediaTypes
from ..utils import gen_user_breadcrumb
from ..compatpatch import ClientCompatPatch


class MediaEndpointsMixin(object):
    """For endpoints in ``/media/``."""

    def media_info(self, media_id):
        """
        Get media info

        :param media_id:
        :return:
        """
        endpoint = 'media/{media_id!s}/info/'.format(**{'media_id': media_id})
        res = self._call_api(endpoint)
        if self.auto_patch:
            [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
             for m in res.get('items', [])]
        return res

    def medias_info(self, media_ids):
        """
        Get multiple media infos

        :param media_ids: list of media ids
        :return:
        """

        if isinstance(media_ids, str):
            media_ids = [media_ids]

        params = {
            '_uuid': self.uuid,
            '_csrftoken': self.csrftoken,
            'media_ids': ','.join(media_ids),
            'ranked_content': 'true',
            'include_inactive_reel': 'true',
        }
        res = self._call_api('media/infos/', query=params)
        if self.auto_patch:
            [ClientCompatPatch.media(m, drop_incompat_keys=self.drop_incompat_keys)
             for m in res.get('items', [])]
        return res

    def media_permalink(self, media_id):
        """
        Get media permalink

        :param media_id:
        :return:
        """
        endpoint = 'media/{media_id!s}/permalink/'.format(**{'media_id': media_id})
        res = self._call_api(endpoint)
        return res

    def media_comments(self, media_id, **kwargs):
        """
        Get media comments. Fixed at 20 comments returned per page.

        :param media_id: Media id
        :param kwargs:
            **max_id**: For pagination
        :return:
        """
        endpoint = 'media/{media_id!s}/comments/'.format(**{'media_id': media_id})
        query = {
            'can_support_threading': 'true'
        }
        if kwargs:
            query.update(kwargs)
        res = self._call_api(endpoint, query=query)

        if self.auto_patch:
            [ClientCompatPatch.comment(c, drop_incompat_keys=self.drop_incompat_keys)
             for c in res.get('comments', [])]
            [ClientCompatPatch.comment(c, drop_incompat_keys=self.drop_incompat_keys)
             for c in res.get('preview_comments', [])]
        return res

    def media_n_comments(self, media_id, n=150, reverse=False, **kwargs):
        """
        Helper method to retrieve n number of comments for a media id

        :param media_id: Media id
        :param n: Minimum number of comments to fetch
        :param reverse: Reverse list of comments (ordered by created_time)
        :param kwargs:
        :return:
        """

        endpoint = 'media/{media_id!s}/comments/'.format(**{'media_id': media_id})

        comments = []
        results = self._call_api(endpoint, query=kwargs)
        comments.extend(results.get('comments', []))

        while (((results.get('has_more_comments') and results.get('next_max_id'))
                or (results.get('has_more_headload_comments') and results.get('next_min_id')))
                and len(comments) < n):

            if results.get('has_more_comments'):
                kwargs.update({'max_id': results.get('next_max_id')})
            else:
                kwargs.update({'min_id': results.get('next_min_id')})

            results = self._call_api(endpoint, query=kwargs)
            comments.extend(results.get('comments', []))
            if not (results.get('next_max_id') or results.get('next_min_id') or results.get('comments')):
                # bail out if no max_id/min_id or comments returned
                break

        if self.auto_patch:
            [ClientCompatPatch.comment(c, drop_incompat_keys=self.drop_incompat_keys)
             for c in comments]

        return sorted(comments, key=lambda k: k['created_at_utc'], reverse=reverse)

    def comment_replies(self, media_id, comment_id, **kwargs):
        """
        Get comment replies. Fixed at 20 replies returned per page.
        Check for 'has_more_tail_child_comments', 'next_max_child_cursor' to determine
        if there are more replies to page through.

        :param media_id: Media id
        :param comment_id: Comment id
        :param kwargs:
            **max_id**: For pagination
        :return:
        """
        endpoint = 'media/{media_id!s}/comments/{comment_id!s}/child_comments/'.format(
            **{'media_id': media_id, 'comment_id': comment_id})
        res = self._call_api(endpoint, query=kwargs)

        if self.auto_patch:
            [ClientCompatPatch.comment(c, drop_incompat_keys=self.drop_incompat_keys)
             for c in res.get('child_comments', [])]
            ClientCompatPatch.comment(res.get('parent_comment'))
        return res

    def comment_inline_replies(self, media_id, comment_id, max_id, **kwargs):
        """
        Get inline comment replies.
        Check for 'next_max_child_cursor' from ``media_comments()``
        to determine if there are inline comment replies to retrieve.

        :param media_id: Media id
        :param comment_id: Comment id
        :param max_id: The comment's 'next_max_child_cursor' value from``media_comments()``
        :return:
        """
        endpoint = 'media/{media_id!s}/comments/{comment_id!s}/inline_child_comments/'.format(
            **{'media_id': media_id, 'comment_id': comment_id})
        query = {'max_id': max_id}
        if kwargs:
            query.update(kwargs)
        res = self._call_api(endpoint, query=query)
        if self.auto_patch:
            [ClientCompatPatch.comment(c, drop_incompat_keys=self.drop_incompat_keys)
             for c in res.get('child_comments', [])]
            ClientCompatPatch.comment(res.get('parent_comment'))
        return res

    def edit_media(self, media_id, caption, usertags=None):
        """
        Edit a media's caption

        :param media_id: Media id
        :param caption: Caption text
        :param usertags: array of user_ids and positions in the format below:

            .. code-block:: javascript

                usertags = [
                    {"user_id":4292127751, "position":[0.625347,0.4384531]}
                ]
        :return:
        """
        if usertags is None:
            usertags = []
        endpoint = 'media/{media_id!s}/edit_media/'.format(**{'media_id': media_id})
        params = {'caption_text': caption}
        params.update(self.authenticated_params)
        if usertags:
            utags = {'in': [{'user_id': u['user_id'], 'position': u['position']} for u in usertags]}
            params['usertags'] = json.dumps(utags, separators=(',', ':'))
        res = self._call_api(endpoint, params=params)
        if self.auto_patch:
            ClientCompatPatch.media(res.get('media'))
        return res

    def delete_media(self, media_id):
        """
        Delete a media

        :param media_id: Media id
        :return:
            .. code-block:: javascript

                {"status": "ok", "did_delete": true}
        """
        endpoint = 'media/{media_id!s}/delete/'.format(**{'media_id': media_id})
        params = {'media_id': media_id}
        params.update(self.authenticated_params)
        return self._call_api(endpoint, params=params)

    def post_comment(self, media_id, comment_text):
        """
        Post a comment.
        Comment text validation according to https://www.instagram.com/developer/endpoints/comments/#post_media_comments

        :param media_id: Media id
        :param comment_text: Comment text
        :return:
            .. code-block:: javascript

                {
                  "comment": {
                    "status": "Active",
                    "media_id": 123456789,
                    "text": ":)",
                    "created_at": 1479453671.0,
                    "user": {
                      "username": "x",
                      "has_anonymous_profile_picture": false,
                      "profile_pic_url": "http://scontent-sit4-1.cdninstagram.com/abc.jpg",
                      "full_name": "x",
                      "pk": 123456789,
                      "is_verified": false,
                      "is_private": false
                    },
                    "content_type": "comment",
                    "created_at_utc": 1479482471,
                    "pk": 17865505612040669,
                    "type": 0
                  },
                  "status": "ok"
                }
        """

        if len(comment_text) > 300:
            raise ValueError('The total length of the comment cannot exceed 300 characters.')
        if re.search(r'[a-z]+', comment_text, re.IGNORECASE) and comment_text == comment_text.upper():
            raise ValueError('The comment cannot consist of all capital letters.')
        if len(re.findall(r'#[^#]+\b', comment_text, re.UNICODE | re.MULTILINE)) > 4:
            raise ValueError('The comment cannot contain more than 4 hashtags.')
        if len(re.findall(r'\bhttps?://\S+\.\S+', comment_text)) > 1:
            raise ValueError('The comment cannot contain more than 1 URL.')

        endpoint = 'media/{media_id!s}/comment/'.format(**{'media_id': media_id})
        params = {
            'comment_text': comment_text,
            'user_breadcrumb': gen_user_breadcrumb(len(comment_text)),
            'idempotence_token': self.generate_uuid(),
            'containermodule': 'comments_feed_timeline',
            'radio_type': self.radio_type,
        }
        params.update(self.authenticated_params)
        res = self._call_api(endpoint, params=params)
        if self.auto_patch:
            ClientCompatPatch.comment(res['comment'], drop_incompat_keys=self.drop_incompat_keys)
        return res

    def delete_comment(self, media_id, comment_id):
        """
        Delete a comment

        :param media_id: Media id
        :param comment_id: Comment id
        :return:
            .. code-block:: javascript

                {"status": "ok"}
        """
        endpoint = 'media/{media_id!s}/comment/{comment_id!s}/delete/'.format(**{
            'media_id': media_id, 'comment_id': comment_id})
        params = {}
        params.update(self.authenticated_params)
        res = self._call_api(endpoint, params=params)
        return res

    def bulk_delete_comments(self, media_id, comment_ids):
        """
        Bulk delete comment

        :param media_id: Media id
        :param comment_ids: List of comment ids
        :return:
            .. code-block:: javascript

                {"status": "ok"}
        """
        if not isinstance(comment_ids, list):
            comment_ids = [comment_ids]
        endpoint = 'media/{media_id!s}/comment/bulk_delete/'.format(**{
            'media_id': media_id})
        params = {
            'comment_ids_to_delete': ','.join(
                [str(comment_id) for comment_id in comment_ids])
        }
        params.update(self.authenticated_params)
        res = self._call_api(endpoint, params=params)
        return res

    def media_likers(self, media_id, **kwargs):
        """
        Get users who have liked a post

        :param media_id:
        :return:
        """
        endpoint = 'media/{media_id!s}/likers/'.format(**{'media_id': media_id})
        res = self._call_api(endpoint, query=kwargs)
        if self.auto_patch:
            [ClientCompatPatch.list_user(u, drop_incompat_keys=self.drop_incompat_keys)
             for u in res.get('users', [])]
        return res

    def media_likers_chrono(self, media_id):
        """
        EXPERIMENTAL ENDPOINT, INADVISABLE TO USE.
        Get users who have liked a post in chronological order

        :param media_id:
        :return:
        """
        warnings.warn('This endpoint is experimental. Do not use.', ClientExperimentalWarning)
        res = self._call_api('media/{media_id!s}/likers_chrono/'.format(**{'media_id': media_id}))
        if self.auto_patch:
            [ClientCompatPatch.list_user(u, drop_incompat_keys=self.drop_incompat_keys)
             for u in res.get('users', [])]
        return res

    def post_like(self, media_id, module_name='feed_timeline'):
        """
        Like a post

        :param media_id: Media id
        :param module_name: Example: 'feed_timeline', 'video_view', 'photo_view'
        :return:
            .. code-block:: javascript

                {"status": "ok"}
        """
        endpoint = 'media/{media_id!s}/like/'.format(**{'media_id': media_id})
        params = {
            'media_id': media_id,
            'module_name': module_name,
            'radio_type': self.radio_type,
        }
        params.update(self.authenticated_params)
        # d query param = flag for double tap
        res = self._call_api(endpoint, params=params, query={'d': '1'})
        return res

    def delete_like(self, media_id, module_name='feed_timeline'):
        """
        Unlike a post

        :param media_id:
        :param module_name: Example: 'feed_timeline', 'video_view', 'photo_view'
        :return:
            .. code-block:: javascript

                {"status": "ok"}
        """
        endpoint = 'media/{media_id!s}/unlike/'.format(**{'media_id': media_id})
        params = {
            'media_id': media_id,
            'module_name': module_name,
            'radio_type': self.radio_type,
        }
        params.update(self.authenticated_params)
        res = self._call_api(endpoint, params=params)
        return res

    def media_seen(self, reels):
        """
        Mark multiple stories as seen

        :param reels: A list of reel media objects, or a dict of media_ids and timings
            as defined below.

            .. code-block:: javascript

                {
                    "1309763051087626108_124317_124317": ["1470355944_1470372029"],
                    "1309764045355643149_124317_124317": ["1470356063_1470372039"],
                    "1309818450243415912_124317_124317": ["1470362548_1470372060"],
                    "1309764653429046112_124317_124317": ["1470356135_1470372049"],
                    "1309209597843679372_124317_124317": ["1470289967_1470372013"]
                }

                where
                    1309763051087626108_124317 = <media_id>,
                    124317 = <media.owner_id>
                    1470355944_1470372029 is <media_created_time>_<view_time>

        :return:
        """
        if isinstance(reels, list):
            # is a list of reel media
            reels_seen = {}
            reels = sorted(reels, key=lambda m: m['taken_at'], reverse=True)
            now = int(time.time())
            for i, reel in enumerate(reels):
                reel_seen_at = now - min(i + 1 + randint(0, 2), max(0, now - reel['taken_at']))
                reels_seen['{0!s}_{1!s}'.format(reel['id'], reel['user']['pk'])] = [
                    '{0!s}_{1!s}'.format(reel['taken_at'], reel_seen_at)]
            params = {'reels': reels_seen}
        else:
            params = {'reels': reels}
        params.update(self.authenticated_params)
        res = self._call_api('media/seen/', params=params, version='v2')
        return res

    def comment_like(self, comment_id):
        """
        Like a comment

        :param comment_id:

        :return:
            .. code-block:: javascript

                {"status": "ok"}
        """
        endpoint = 'media/{comment_id!s}/comment_like/'.format(**{'comment_id': comment_id})
        params = self.authenticated_params
        return self._call_api(endpoint, params=params)

    def comment_likers(self, comment_id):
        """
        Get users who have liked a comment

        :param comment_id:
        :return:
        """
        endpoint = 'media/{comment_id!s}/comment_likers/'.format(**{'comment_id': comment_id})
        res = self._call_api(endpoint)
        if self.auto_patch:
            [ClientCompatPatch.list_user(u, drop_incompat_keys=self.drop_incompat_keys)
             for u in res.get('users', [])]
        return res

    def comment_unlike(self, comment_id):
        """
        Unlike a comment

        :param comment_id:
        :return:
            .. code-block:: javascript

                {"status": "ok"}
        """
        endpoint = 'media/{comment_id!s}/comment_unlike/'.format(**{'comment_id': comment_id})
        params = self.authenticated_params
        return self._call_api(endpoint, params=params)

    def save_photo(self, media_id, added_collection_ids=None):
        """
        Save a photo

        :param media_id: Media id
        :param added_collection_ids: optional list of collection IDs to add the media to
        :return:
            .. code-block:: javascript

                {"status": "ok"}
        """
        endpoint = 'media/{media_id!s}/save/'.format(**{'media_id': media_id})
        params = {'radio_type': self.radio_type}
        if added_collection_ids:
            if isinstance(added_collection_ids, str):
                added_collection_ids = [added_collection_ids]
            params['added_collection_ids'] = json.dumps(added_collection_ids, separators=(',', ':'))
        params.update(self.authenticated_params)
        return self._call_api(endpoint, params=params)

    def unsave_photo(self, media_id, removed_collection_ids=None):
        """
        Unsave a photo

        :param media_id:
        :param removed_collection_ids: optional list of collection IDs to remove the media from
        :return:
            .. code-block:: javascript

                {"status": "ok"}
        """
        endpoint = 'media/{media_id!s}/unsave/'.format(**{'media_id': media_id})
        params = {'radio_type': self.radio_type}
        if removed_collection_ids:
            if isinstance(removed_collection_ids, str):
                removed_collection_ids = [removed_collection_ids]
            params['removed_collection_ids'] = json.dumps(removed_collection_ids, separators=(',', ':'))
        params.update(self.authenticated_params)
        return self._call_api(endpoint, params=params)

    def disable_comments(self, media_id):
        """
        Disable comments for a media

        :param media_id:
        :return:
            .. code-block:: javascript

                {"status": "ok"}
        """
        endpoint = 'media/{media_id!s}/disable_comments/'.format(**{'media_id': media_id})
        params = {
            '_csrftoken': self.csrftoken,
            '_uuid': self.uuid,
        }
        res = self._call_api(endpoint, params=params, unsigned=True)
        return res

    def enable_comments(self, media_id):
        """
        Enable comments for a media

        :param media_id:
        :return:
            .. code-block:: javascript

                {"status": "ok"}
        """

        endpoint = 'media/{media_id!s}/enable_comments/'.format(**{'media_id': media_id})
        params = {
            '_csrftoken': self.csrftoken,
            '_uuid': self.uuid,
        }
        res = self._call_api(endpoint, params=params, unsigned=True)
        return res

    def media_only_me(self, media_id, media_type, undo=False):
        """
        Archive/unarchive a media so that it is only viewable by the owner.

        :param media_id:
        :param media_type: One of :attr:`MediaTypes.PHOTO`, :attr:`MediaTypes.VIDEO`, or :attr:`MediaTypes.CAROUSEL`
        :param undo: bool

        :return:
            .. code-block:: javascript

                {"status": "ok"}
        """
        if media_type not in MediaTypes.ALL:
            raise ValueError('Invalid media type.')

        endpoint = 'media/{media_id!s}/{only_me!s}/'.format(**{
            'media_id': media_id,
            'only_me': 'only_me' if not undo else 'undo_only_me'
        })
        params = {'media_id': media_id}
        params.update(self.authenticated_params)
        res = self._call_api(endpoint, params=params, query={'media_type': media_type})
        return res

    def media_undo_only_me(self, media_id, media_type):
        """
        Undo making a media only me.

        :param media_id:
        :param media_type: One of :attr:`MediaTypes.PHOTO`, :attr:`MediaTypes.VIDEO`, or :attr:`MediaTypes.CAROUSEL`
        """
        return self.media_only_me(media_id, media_type, undo=True)

    def story_viewers(self, story_pk, **kwargs):
        """
        Get list of story viewers

        :param story_pk: Story media's PK identifier, e.g. "1700000123"
        :param kwargs:
            **max_id**: For pagination
        :return:
        """
        endpoint = 'media/{story_pk!s}/list_reel_media_viewer/'.format(
            story_pk=story_pk)
        return self._call_api(endpoint, query=kwargs)



import json


class HighlightsEndpointsMixin(object):
    """For endpoints in ``/highlights/`` or related to the highlights feature."""

    def stories_archive(self, **kwargs):
        """
        Returns the authenticated user's story archive. The returned items's id
        value is passed to ``reels_media()`` to retrieve

        Example:
            .. code-block:: python

                archived_stories = api.stories_archive()
                if archived_stories.get('items):
                    item_ids = [a['id'] for a in archived_stories['items']]
                    archived_stories_media = api.reels_media(user_ids=item_ids)

        :return:
            .. code-block:: javascript

                {
                    "items": [{
                        "timestamp": 1510090000,
                        "media_count": 3,
                        "id": "archiveDay:1710000000",
                        "reel_type": "archive_day_reel",
                        "latest_reel_media": 1510090000
                    }],
                    "num_results": 1,
                    "more_available": false,
                    "max_id": null,
                    "status": "ok"
                }
        """
        query = {'include_cover': '0'}
        if kwargs:
            query.update(kwargs)
        return self._call_api('archive/reel/day_shells/', query=query)

    def highlights_user_feed(self, user_id):
        """
        Returns a user's highlight tray

        :param user_id:
        """
        endpoint = 'highlights/{user_id!s}/highlights_tray/'.format(user_id=user_id)
        return self._call_api(endpoint)

    def highlight_create(
            self, media_ids, cover_media_id=None,
            title='Highlights', source='self_profile'):
        """
        Create a new highlight

        :param media_ids: A list of media_ids
        :param cover_media_id: The media_id for the highlight cover image
        :param title: Title of the highlight
        :param module: The UI module via which the highlight is created
        """
        if not (media_ids and isinstance(media_ids, list)):
            raise ValueError('media_ids must be a non-empty list')

        if not cover_media_id:
            cover_media_id = media_ids[0]

        if not title:
            title = 'Highlights'

        if len(title) > 16:
            raise ValueError('title must not exceed 16 characters')

        cover = {
            'media_id': cover_media_id,
            'crop_rect': json.dumps(
                [0.0, 0.21830457, 1.0, 0.78094524], separators=(',', ':'))
        }
        params = {
            'media_ids': json.dumps(media_ids, separators=(',', ':')),
            'cover': json.dumps(cover, separators=(',', ':')),
            'source': source,
            'title': title,
        }
        params.update(self.authenticated_params)
        return self._call_api('highlights/create_reel/', params=params)

    def highlight_edit(
            self, highlight_id, cover_media_id=None,
            added_media_ids=[], removed_media_ids=[],
            title=None, source='story_viewer'):
        """
        Edits a highlight

        :param highlight_id: highlight_id, example 'highlight:1770000'
        :param cover_media_id: The media_id for the highlight cover image
        :param added_media_ids: List of media_id to be added
        :param removed_media_ids: List of media_id to be removed
        :param title: Title of the highlight
        :param module: The UI module via which the highlight is created
        """
        endpoint = 'highlights/{highlight_id!s}/edit_reel/'.format(
            highlight_id=highlight_id
        )

        # sanitise inputs
        if not added_media_ids:
            added_media_ids = []
        elif not isinstance(added_media_ids, list):
            raise ValueError('added_media_ids must be a list')

        if not removed_media_ids:
            removed_media_ids = []
        elif not isinstance(removed_media_ids, list):
            raise ValueError('removed_media_ids must be a list')

        if title and len(title) > 16:
            raise ValueError('title must not exceed 16 characters')

        if not (added_media_ids or removed_media_ids or cover_media_id or title):
            raise ValueError('No edited values')

        params = {
            'added_media_ids': json.dumps(added_media_ids, separators=(',', ':')),
            'removed_media_ids': json.dumps(removed_media_ids, separators=(',', ':')),
            'source': source,
        }
        if title:
            params['title'] = title
        if cover_media_id:
            cover = {
                'media_id': cover_media_id,
                'crop_rect': json.dumps(
                    [0.0, 0.21830457, 1.0, 0.78094524], separators=(',', ':'))
            }
            params['cover'] = json.dumps(cover, separators=(',', ':'))

        params.update(self.authenticated_params)
        return self._call_api(endpoint, params=params)

    def highlight_delete(self, highlight_id):
        """
        Deletes specified highlight

        :param highlight_id: highlight_id, example 'highlight:1770000'
        """
        endpoint = 'highlights/{highlight_id!s}/delete_reel/'.format(
            highlight_id=highlight_id
        )
        return self._call_api(endpoint, params=self.authenticated_params)
